"""
File Purpose: QuantityLoader, the base class for quantity loaders.

Note: "var" refers to quantity name; "quant" refers to the quantity itself.


Note: MetaQuantTracking predefines some things for tracking calcs, directly in the class namespace.
    E.g., known_var, known_pattern, known_setter.
In case anyone is looking through the codebase to understand these functions,
    this file is the one that defines them, though it doesn't use the usual 'def' notation,
    so, here are some strings that will pop up for searches like that, instead.
        def known_var(...)      # see MetaQuantTracking
        def known_pattern(...)  # see MetaQuantTracking
        def known_setter(...)   # see MetaQuantTracking
    if you landed here by doing command+F on the codebase, see MetaQuantTracking for details.
    (see also: quantity_tools.py, which defines the classes which these are instances of,
        i.e. DecoratingVars, DecoratingPatterns, and DecoratingSetters.)
"""
import builtins
import warnings

import xarray as xr

from .caching import VarCacheSingles
from .quant_tree import QuantTree
from .quantity_tools import (
    MatchedQuantity, MatchedVar, MatchedPattern,
    Pattern,
    DecoratingVars, DecoratingPatterns, DecoratingSetters,
    CallDepthMixin,
)
from ..dimensions import BehaviorHaver, MetaBehaviorHaver
from ..errors import (
    InputError, InputConflictError,
    QuantCalcError, OverrideNotApplicableError, CacheNotApplicableError,
    FormulaMissingError, LoadingNotImplementedError,
    SetvarNotImplementedError, TypevarNanError,
    DimensionalityError,
)
from ..tools import (
    alias, simple_property,
    using_attrs, maintaining_attrs,
    RESULT_MISSING, UNSET,
    xarray_assign, xarray_scale_coords,
    copy_via_pickle,
)
from ..defaults import DEFAULTS


''' --------------------- MetaQuantTracking --------------------- '''

class MetaQuantTracking(MetaBehaviorHaver):
    '''metaclass which predefines some things for tracking calcs, in the class namespace:
        KNOWN_VARS - dict of {var: LoadableVar} for all functions decorated with @known_var
        KNOWN_PATTERNS - dict of {pattern: LoadablePattern} for all functions decorated with @known_pattern.
        KNOWN_SETTERS - dict of {var: f} for all functions decorated with @known_setter
        known_var - instance of DecoratingVars; use @known_var or @known_var(name=varname) to decorate functions.
            use this for vars, which require exact match for name.
            E.g. name='n' for number density.
            (Note, if function.__name__ is 'get_{var}', name='{var}' is implied, if not provided explicitly.)
        known_pattern - instance of DecoratingPatterns; use @known_pattern(pattern) to decorate functions.
            use this for patterns, which require regex match for name.
            E.g. pattern='mean_(.+)' for mean of any var.
            decorated functions will be supplied with kwarg _match=re.fullmatch(pattern, var).
        known_setter - instance of DecoratingSetters; use @known_setter or @known_setter(name=varname).
            use this for setters, which are functions that set the value of vars.
            For more details see setvars or set_var.

        Also, copy base classes' KNOWN_VARS, KNOWN_PATTERNS, and KNOWN_SETTERS to this class, if they exist.

    Note that __prepare__ gets called before class definition begins, and fills class namespace,
    then __init__ gets called after class definition ends. For example:
        class MySubclass(metaclass=MetaQuantTracking):
            # <-- __prepare__ runs here. Roughly: locals().update(__prepare__())
            pass  # or, define other functions, variables, etc, as desired.
        # <-- __init__ runs here. Roughly: MetaQuantTracking.__init__(MySubclass)
    '''
    @classmethod
    def __prepare__(_metacls, _name, bases):
        super_result = super().__prepare__(_name, bases)
        KNOWN_VARS = dict()  # [TODO] use ordered dict for earlier python (before dict order guaranteed)?
        KNOWN_PATTERNS = dict()
        KNOWN_SETTERS = dict()
        for base in bases:
            # So that keys appear in __mro__ order, must iterate through bases in order.
            # So, the first time a key appears, it points to the earliest (in __mro__ order)
            # implementation, so it should not be overwritten if it appears again.
            #    Hence, avoid dict.update; use dict.setdefault instead.
            #
            # Additionally, need to avoid the following:
            #    'vv' defined in Class0.
            #    Class1 subclasses Class0 but does NOT also define 'vv'.
            #    Class2 subclasses Class0 and defines 'vv' too.
            #    class Class3(Class1, Class2): ...
            #    now, Class3.mro() has [Class1, Class2, Class0], so:
            #      Class3.get_vv uses Class2.get_vv.
            #        (This is good; should use the earliest-in-mro() get_vv to get value.)
            #      Class3.KNOWN_VARS['vv'] uses Class1.KNOWN_VARS['vv'].
            #        (This is bad; should use the KNOWN_VARS attached to get_vv.)
            # To avoid this, below checks where key was defined if it appears in dict already;
            #    if it was defined in a subclass, don't overwrite.
            base_kvars = getattr(base, 'KNOWN_VARS', {})
            for k, v in base_kvars.items():
                try:
                    existing = KNOWN_VARS[k]
                except KeyError:  # didn't exist yet
                    KNOWN_VARS[k] = v
                else:  # existed already
                    # only overwrite if this base is a subclass of where the existing value was defined.
                    if issubclass(base, existing.cls_where_defined):
                        KNOWN_VARS[k] = v
            base_kpats = getattr(base, 'KNOWN_PATTERNS', {})
            for k, v in base_kpats.items():
                try:
                    existing = KNOWN_PATTERNS[k]
                except KeyError:
                    KNOWN_PATTERNS[k] = v
                else:
                    if issubclass(base, existing.cls_where_defined):
                        KNOWN_PATTERNS[k] = v
            base_ksetters = getattr(base, 'KNOWN_SETTERS', {})
            for k, v in base_ksetters.items():
                try:
                    existing = KNOWN_SETTERS[k]
                except KeyError:
                    KNOWN_SETTERS[k] = v
                else:
                    if issubclass(base, existing.cls_where_defined):
                        KNOWN_SETTERS[k] = v
        known_var = DecoratingVars(KNOWN_VARS)
        known_pattern = DecoratingPatterns(KNOWN_PATTERNS)
        known_setter = DecoratingSetters(KNOWN_SETTERS)
        return dict(**super_result,
                    KNOWN_VARS=KNOWN_VARS, KNOWN_PATTERNS=KNOWN_PATTERNS, KNOWN_SETTERS=KNOWN_SETTERS,
                    known_var=known_var, known_pattern=known_pattern, known_setter=known_setter)

    def __init__(cls, *args, **kw):
        cls.known_var.cls_associated_with = cls
        for k, v in cls.KNOWN_VARS.items():
            if not hasattr(v, 'cls_where_defined'):
                v.cls_where_defined = cls
        cls.known_pattern.cls_associated_with = cls
        for k, v in cls.KNOWN_PATTERNS.items():
            if not hasattr(v, 'cls_where_defined'):
                v.cls_where_defined = cls
        cls.known_setter.cls_associated_with = cls
        for k, v in cls.KNOWN_SETTERS.items():
            if not hasattr(v, 'cls_where_defined'):
                v.cls_where_defined = cls
        super().__init__(*args, **kw)


''' --------------------- QuantityLoader --------------------- '''

class QuantityLoader(BehaviorHaver, CallDepthMixin, metaclass=MetaQuantTracking):
    '''base class for quantity loaders. See self.get or self.__call__ for details.'''
    setvars = simple_property('_setvars', setdefault=VarCacheSingles,
            doc='''VarCache of vars set via self.set_var(). Returns these values when appropriate,
            i.e. whenever self.behavior is compatible with the behavior in the cache.
            To empty the cache, use self.setvars.clear() to empty the cache.''')

    get = alias('__call__')

    def __call__(self, var, *args, name=UNSET, item=False, verbose=UNSET, **kw):
        '''returns value of var from self.
        result is probably an xarray.DataArray, but not guaranteed.
        
        var: str or iterable of strs.
            Name of the var(s) to load. E.g. 'n' for number density, or ['n', 'u'] for number density & velocity.
            If multiple vars: returns an xarray.Dataset of all vars, via self.get_vars.

            Determine how to load each var, as follows:
                - (caching) if var in self.cache, with matching self.behavior_attrs, use value from cache.
                    [TODO] - caching not yet implemented. May allow for better efficiency. 
                - (setvars) if var in self.setvars, with matching self.behavior_attrs, use value from setvars.
                    [TODO] - improve set_var functionality.
                    set_var will allow user to apply PlasmaCalcs calculations to arbitrary values,
                    not just values from one of the hookups. Useful for testing & quick calculations.
                - (KNOWN_VARS) if var in self.KNOWN_VARS,
                    use the corresponding function to get it.
                - (KNOWN_PATTERNS) if var matches a pattern from self.KNOWN_PATTERNS,
                    use the corresponding function to get it.
                - (direct) attempt to load var "directly", via self.load_direct.
                    load_direct will almost always end up loading values directly from a file (e.g., "data").
                    However, there is one more chance for it to get intercepted: via direct_overrides.
                    - (direct_overrides) if var in self.direct_overrides, attempts self.direct_overrides[var].
                        The idea is to use direct_overrides when requiring alternate instructions for
                        what would otherwise be a "base" var.
                        E.g., 'n' may be a "base" var, but if quasineutral then 'ne' is not saved in a file;
                            so, base_overrides[<key for density>] may tell how to get 'ne'.
                        Note: for overrides which depend on current state, use direct_overrides_dynamic.
                        For overrides which are "always" implemented (not toggled by other things), use direct_overrides.
                    - (fromfile) load_direct uses self.load_fromfile whenever direct_overrides is not applicable.

            Those are checked in the order listed.
            If none of those work, raise FormulaMissingError.

        name: UNSET, None, or str
            try to set result.name = name.
            UNSET --> use name = var.
        item: bool
            if True, convert result to single value (e.g., python float) via result.item().
            This will cause crash if result is not a single value;
                it will also cause all metadata stored in the result to be lost.
        verbose: UNSET, bool, or int
            set self.verbose during this call to self.
            UNSET --> use self.verbose (unchanged)

        kw may additionally contain any keys from self.kw_call_options().
            if it does, pop those values, and temporarily set the corresponding attr.
            E.g.: self('n', units='si', fluid=1)
                --> temporarily set units='si', fluid=1, while getting 'n'.
            See self.help_call_options() for more details.

        [EFF] passes _match=re.fullmatch(pattern, var) to the getter function,
            if the match is from KNOWN_PATTERNS (but not if it is from KNOWN_VARS).
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        # handle multiple vars
        if not isinstance(var, str):
            return self.get_vars(var, *args, name=name, item=item, verbose=verbose, **kw)
        # pop kw
        using = self._pop_kw_call_options(kw)
        if verbose is not UNSET: using['verbose'] = verbose
        # load the value of var
        with self._increment_call_depth():
            result = RESULT_MISSING
            with self.using(**using):
                # verbose / debugging
                if DEFAULTS.DEBUG >= 7 or getattr(self, 'verbose', False) >= 2:
                    depth = self.call_depth
                    print('| '*depth + f'(call_depth={depth}) get var={var!r}')
                with self.using(_inside_quantity_loader_call_logic=True):
                    if result is RESULT_MISSING:
                        # try to get from cache
                        try:
                            result = self.get_set_or_cached(var)  # [TODO] get_set xarray or simple value here;
                        except CacheNotApplicableError:
                            pass
                    if result is RESULT_MISSING:
                        # try to get from KNOWN_VARS or KNOWN_PATTERNS
                        try:
                            matched_quantity = self.match_var(var)
                        except FormulaMissingError as err0:  # (direct)   # no match in KNOWN_VARS & KNOWN_PATTERNS.
                            # >> actually get value from load_direct. Crash if that fails <<
                            try:
                                result = self.load_direct(var, *args, **kw)  # [TODO] get_set array value here.
                            except Exception as err1:
                                raise err1 from err0   # traceback will include info from err0 and err1.
                if result is RESULT_MISSING:  # (important: outside of using(_inside_quantity_loader_call_logic=True))!
                    # >> actually get value from KNOWN_VARS or KNOWN_PATTERNS <<
                    result = matched_quantity.load_value(self, *args, **kw)
                assert result is not RESULT_MISSING, 'if this assertion fails, there is a coding error above.'
                # "post-processing" of result (intentionally still inside self.using(...))
                result = self._call_postprocess(result, var=var, name=name, item=item)
        return result

    def _pop_kw_call_options(self, kw):
        '''pop all self.kw_call_options() from kw, returning dict of popped options.'''
        options = self.kw_call_options()
        return {key: kw.pop(key) for key in tuple(kw.keys()) if key in options}

    def _call_postprocess(self, result, *, var, name=UNSET, item=UNSET):
        '''postprocess result from self.__call__. Called during self.__call__.
        (self.call_depth inside here will tell depth of the current call; depth=1 for top level.)

        result: any value, probably an xarray.DataArray
            result from self.__call__, before postprocessing.
        var, name, item: UNSET or value
            passed directly from self.__call__.

        The implementation here does the following (subclasses might override / add to this):
            (1) if self.verbose >= 4, print a message about getting var.
            (2) result = self.attach_extra_coords(result).
            (3) if name was provided, set result.name = name, if possible.
            (4) if self.assign_behavior_attrs at this call depth, do so now.
            (5) if self.call_depth == 1, call self._call_postprocess_toplevel.
            (6) if item, convert result to single value via result.item().
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        depth = self.call_depth
        # (1)
        if getattr(self, 'verbose', False) >= 4:
            print('/ '*depth + f'(call_depth={depth}) got var={var!r}')
        # (2)
        result = self.attach_extra_coords(result)
        # (3)
        if name is UNSET:
            name = var
        try:
            result.name = name
        except AttributeError:  # e.g. 'float' object has no attribute 'name'
            if DEFAULTS.DEBUG > 2:
                raise
            # else, pass; it's fine, just don't name the result.
        # (4)
        assign_mode = self.assign_behavior_attrs
        if assign_mode:
            maxdepth = self.assign_behavior_attrs_max_call_depth
            if (maxdepth is None) or self.call_depth <= maxdepth:
                if assign_mode == True or assign_mode == 'nondefault':
                    include_xr = not self.assign_behavior_attrs_skip_xr
                    result = self.behavior.assign_nondefault_attrs(result, ql=self, include_xr=include_xr)
                else:
                    assert (assign_mode == 'all'), 'else, coding error....'
                    result = self.behavior.assign_attrs(result)
        # (5)
        if self.call_depth == 1:
            result = self._call_postprocess_toplevel(result, var=var, name=name, item=item)
        # (6)
        if item:
            if result.size != 1:
                errmsg = (f'expected result.size==1 when item=True, but got size={result.size}\n'
                          f'This error is likely caused by having a dim with length>1 in self.dims,\n'
                          f'for one of the dims used when getting var={var!r}.')
                raise DimensionalityError(errmsg)
            result = result.item()
        return result

    def _call_postprocess_toplevel(self, result, *, var, name=UNSET, item=UNSET):
        '''additional postprocessing for self.__call__ when call_depth=1.
        called from self._call_postprocess, after doing other postprocessing, when call_depth=1.

        result: any value, probably an xarray.DataArray
            result from self.__call__, after other postprocessing (except `item`).
        var, name, item: UNSET or value
            passed directly from self.__call__.
            Don't need to handle these here because self._call_postprocess will handle it.

        The implementation here does the following (subclasses might override / add to this):
            (1) self._apply_toplevel_scale_coords (does nothing if self.toplevel_scale_coords is empty)
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        # (1)
        result = self._apply_toplevel_scale_coords(result)
        return result

    @property
    def _extra_kw_for_quantity_loader_call(self):
        '''extra kwargs which can be used to set attrs self during self.__call__.
        (This is in addition to all self.behavior_attrs().).
        The implementation here returns:
            ['enable_fromfile', 'typevar_crash_if_nan',
            'assign_behavior_attrs', 'assign_behavior_attrs_max_call_depth',
            'assign_behavior_attrs_skip_xr'].
        '''
        result = [
            'enable_fromfile',
            'typevar_crash_if_nan',
            'assign_behavior_attrs',
            'assign_behavior_attrs_max_call_depth',
            'assign_behavior_attrs_skip_xr',
        ]
        return result

    enable_fromfile = simple_property('_enable_fromfile', default=True,
            doc='''bool: whether self.load_fromfile is enabled during self.load_direct.
            If False, raise QuantCalcError if load_direct can't get value without load_fromfile().''')

    assign_behavior_attrs = simple_property('_assign_behavior_attrs', default=False,
            valid=[False, True, 'nondefault', 'all'],
            doc='''whether to assign self.behavior values as attrs of result when calling self.
            False --> don't use self.behavior code architecture to assign attrs.
            True --> equivalent to 'nondefault'
            'nondefault' --> self.behavior.assign_nondefault_attrs(result)
                        (for brevity, it does not assign behavior attrs with "default" value.)
            'all' --> self.behavior.assign_attrs(result).

            [EFF] only assigns attrs at call_depth >= self.assign_behavior_attrs_max_call_depth.
                (default: only assigns attrs at call_depth=1, i.e. at top level.''')

    assign_behavior_attrs_max_call_depth = simple_property('_assign_behavior_attrs_min_call_depth',
            default=1,
            doc='''max call_depth at which to assign_behavior_attrs to result,
            if self.assign_behavior_attrs indicates to assign behavior attrs.
            default 1, i.e. only assign if at top level.
            Use None to indicate "no max detph".''')

    assign_behavior_attrs_skip_xr = simple_property('_assign_behavior_attrs_skip_xr',
            default=False,
            doc='''whether to use include_xr=False if self.assign_behavior_attrs,
            during self.behavior.assign_nondefault_attrs.
            Use this if you want to assign behavior attrs EXCEPT array-valued behavior attrs.''')

    def kw_call_options(self, *, sorted=True):
        '''returns list of kwarg names which can be used to set attrs self during self.__call__.
        (see self.__call__ for more details).
        Here, returns list(self.behavior_attrs) + list(self._extra_kw_for_quantity_loader_call)
        '''
        battrs = [str(battr) for battr in self.behavior_attrs]
        result = battrs + list(self._extra_kw_for_quantity_loader_call)
        return builtins.sorted(result) if sorted else result

    def get_vars(self, vars, *args, return_type='dataset', missing_vars=UNSET, **kw):
        '''returns values of vars from self.
        result is probably an xarray.Dataset, but not guaranteed; also depends on return_type.

        Equivalent to self(vars, *args, return_type='dataset', **kw).
            (Actually, self(vars, ...) will call self.get_vars(vars, ...).)

        vars: iterable of strs
            Names of the vars to load. ['n', 'u'] for number density & velocity.
            if any of these vars returns a return_type object, expand its keys,
                e.g. if 'myDSvar' returns dataset with 'myvar1', 'myvar2',
                then ['n', 'myDSvar'] gives dataset with 'n', 'myvar1', 'myvar2'.
        return_type: 'dataset' or 'dict'
            if 'dataset', return result as xarray.Dataset.
                the data_var names will be the same as the var names.
            if 'dict', return result as dict of {var: value}.
        missing_vars: UNSET, 'ignore', 'warn', or 'raise'
            what to do if any vars cause FormulaMissingError.
            UNSET --> use self.missing_vars if it exists, else 'raise'.
            'ignore' --> ignore missing vars, and don't include them in the result.
            'warn' --> ignore missing vars, but print a warning.
            'raise' --> raise FormulaMissingError if any vars are missing.

        additional args & kwargs are passed to self(...).
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if missing_vars is UNSET: missing_vars = getattr(self, 'missing_vars', 'raise')
        result = xr.Dataset() if return_type == 'dataset' else dict()
        for var in vars:
            value = self._get_maybe_missing_var(var, *args, missing_vars=missing_vars, **kw)
            if value is not None:
                is_return_type = (return_type == 'dataset' and isinstance(value, xr.Dataset)) \
                                 or (return_type == 'dict' and isinstance(value, dict))
                if is_return_type:
                    for k, v in value.items():
                        result[k] = v
                else:
                    result[var] = value
        return result

    def _get_maybe_missing_var(self, var, *args, missing_vars=UNSET, **kw):
        '''return value of var, or None if FormulaMissingError and missing_vars 'ignore' or 'warn'.
        missing_vars: UNSET, 'ignore', 'warn', or 'raise'
            what to do if any var causes FormulaMissingError.
            UNSET --> use self.missing_vars if it exists, else 'raise'.
            'ignore' --> return None.
            'warn' --> return None, but also print a warning.
            'raise' --> raise FormulaMissingError.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if missing_vars is UNSET:
            missing_vars = getattr(self, 'missing_vars', 'raise')
        try:
            return self(var, *args, **kw)
        except FormulaMissingError as err:
            if missing_vars == 'ignore':
                return None
            elif missing_vars == 'warn':
                warnings.warn(f'FormulaMissingError for var={var!r}, when missing_vars="warn".')
                return None
            else:
                raise err

    def _provided_val(self, var, _val=None, _known_vals=dict()):
        '''returns the value of var, either from _known_vals or _val.
        if _val provided, return it; if '_{var}' in _known_vals, return it;
        if both provided, crash with InputConflictError (unless they are the same object),
        else, return None.

        Can use this internally to avoid redundant recalculations. (See e.g. VectorArithmeticLoader)

        '''
        _vstr = f'_{var}'
        if _vstr in _known_vals and (_val is not None) and (_val is not _known_vals[_Astr]):
            raise InputConflictError(f'cannot provide both _val and {_vstr}')
        return _known_vals.get(_vstr, _val)

    # # # ATTRS MANAGEMENT # # #
    using_attrs = using_attrs
    using = alias('using_attrs')
    maintaining_attrs = maintaining_attrs
    maintaining = alias('maintaining_attrs')

    # # # LOAD FROM KNOWN_VARS OR KNOWN_PATTERNS # # #
    quant_tree_cls = QuantTree   # class to use when making quant_tree from self & var.
    matched_var_cls = MatchedVar   # class to use for MatchedVar in self.match_var
    matched_pattern_cls = MatchedPattern   # class to use for MatchedPattern in self.match_var

    @classmethod
    def match_var(cls, var, *, check=['KNOWN_VARS', 'KNOWN_PATTERNS']):
        '''match var from cls.KNOWN_VARS or cls.KNOWN_PATTERNS, or raise FormulaMissingError.
        
        returns result=MatchedQuantity(var, loadable, _match=_match) where:
            loadable is the LoadableQuantity associated with this var,
            _match is:
                None, if var in cls.KNOWN_VARS;
                re.fullmatch(pattern, var), if var matches any pattern in cls.KNOWN_PATTERNS.
                    if var matches multiple patterns, only the first matching pattern is used.

            Uses MatchedVar if match from KNOWN_VARS, MatchedPattern if from KNOWN_PATTERNS.
            (note that both MatchedVar and MatchedPattern subclass MatchedQuantity.)
        
        check: str or list of str from ['KNOWN_VARS', 'KNOWN_PATTERNS']
            where to check for matches. Default is to check KNOWN_VARS and KNOWN_PATTERNS.
            E.g. to only check KNOWN_PATTERNS, use check=['KNOWN_PATTERNS'].

        loadable and _match can be retrieved via result.loadable and result._match.
        '''
        if isinstance(check, str):
            check = [check]
        if any(c not in ['KNOWN_VARS', 'KNOWN_PATTERNS'] for c in check):
            raise InputError(f"invalid check={check!r}; expected only vals from ['KNOWN_VARS', 'KNOWN_PATTERNS']")
        if len(check) == 0:
            raise InputError("check=[]; expected at least one value from ['KNOWN_VARS', 'KNOWN_PATTERNS']")
        # match in KNOWN_VARS
        if 'KNOWN_VARS' in check:
            try:
                lquant = cls.KNOWN_VARS[var]
            except KeyError:
                pass
            else:
                return cls.matched_var_cls(var, lquant, _match=None)
        # match in KNOWN_PATTERNS
        if 'KNOWN_PATTERNS' in check:
            for pattern, lquant in cls.KNOWN_PATTERNS.items():
                _match = pattern.fullmatch(var)
                if _match:
                    return cls.matched_pattern_cls(var, lquant, _match=_match)
        # didn't match.
        errmsg = f"var={var!r}; no match in {cls.__name__}." + ' or .'.join(check)
        raise FormulaMissingError(errmsg)

    def has_var(cls, var):
        '''return whether self can load var. True if self.match_var(var) is found, else False.
        Subclasses might override, to include checks for whether var can be loaded from data.
        [TODO] also check if var in self.cache or self.setvars.
        '''
        try:
            cls.match_var(var)
        except FormulaMissingError:
            return False
        else:
            return True

    def match_var_tree(self=UNSET, var=UNSET, **kw_quant_tree_from_quantity_loader):
        '''return QuantTree of MatchedQuantity objects from matching var and all dependencies,
        using self.KNOWN_VARS and self.KNOWN_PATTERNS when searching for matches.

        var must be provided; var=UNSET will raise an error (helpful if tried calling this as a classmethod).

        See also: type(self).cls_var_tree, for the classmethod version of this function.
            Most of the time it is possible to get tree without any details from self,
            but sometimes not. e.g. when getting collision frequencies, self.fluid affects deps.

        additional kwargs will be passed to QuantTree.from_quantity_loader(...),
            which passes kwargs from self.kw_call_options() into self.using(**kw) while getting deps.
        '''
        if not isinstance(self, QuantityLoader):
            errmsg = (f'Expected QuantityLoader self, got type(self)={type(self)}.\n'
                      'This might occur if you called cls.match_var_tree, instead of obj.match_var_tree.\n'
                      'Use cls.cls_var_tree() instead of cls.match_var_tree(), to try getting tree from class.\n'
                      '  It should succeed for most vars, but fail for vars whose tree depends on present values.')
            raise InputError(errmsg)
        if var is UNSET:
            raise TypeError("match_var_tree() missing 1 required positional argument: 'var'")
        return self.quant_tree_cls.from_quantity_loader(self, var, **kw_quant_tree_from_quantity_loader)

    quant_tree = match_var_tree   # [TODO] use alias instead? (But then it's inaccessible from cls.quant_tree.)
    tree = match_var_tree

    @classmethod
    def cls_var_tree(cls, var, *, missing_ok=False):
        '''return QuantTree of MatchedQuantity objects from matching var and all dependencies,
        using self.KNOWN_VARS and self.KNOWN_PATTERNS when searching for matches.
        missing_ok: bool
            whether to be lenient sometimes when missing details that would allow to fully determine deps.
            see help(MatchedQuantity.dep_vars) for more details.
        '''
        return cls.quant_tree_cls.from_quantity_loader(cls, var, missing_ok=missing_ok)

    def match_var_loading_dims(self, var, **kw_loading_dims):
        '''return dims for loading var across.
        Result will probably vary across these dims (but not guaranteed, if any dependency uses reduces_dims.)
        These are all Dimension dims, not maindims. (E.g. 'fluid' and 'snap', but not 'x', 'y', 'z').

        Equivalent: self.match_var_tree(var).loading_dims(**kw_loading_dims)
        '''
        tree = self.match_var_tree(var)  # in its own line for easier debugging in case of crash.
        return tree.loading_dims(**kw_loading_dims)

    def match_var_result_dims(self, var, **kw_result_dims):
        '''return dims which result of cls(var) will vary across.
        These are all Dimension dims, not maindims. (E.g. 'fluid' and 'snap', but not 'x', 'y', 'z').

        Equivalent: cls.match_var_tree(var).result_dims(**kw_result_dims)
        '''
        tree = self.match_var_tree(var)  # in its own line for easier debugging in case of crash.
        return tree.result_dims(**kw_result_dims)

    def match_var_result_size(self, var, *, maindims=True, **kw_result_dims):
        '''return size (number of elements) which self(var) will have.
        (Efficient; doesn't actually get self(var).)
        Depends on current values of relevant dims. (E.g., self.fluid, not self.fluids)

        maindims: bool
            if True, include maindims_shape when calculating size.
        '''
        result_dims = self.match_var_result_dims(var, **kw_result_dims)
        n_dimpoints = self.current_n_dimpoints(dims=result_dims)
        if maindims:
            maindims_size = getattr(self, 'maindims_size', 1)
            size = n_dimpoints * maindims_size
        else:
            size = n_dimpoints
        return size
        
    # # # LOAD DIRECT # # #
    def load_direct(self, var, *args, **kw):
        '''load var "directly", from some source which is not known by the main part of PlasmaCalcs.
        Attempt the following, returning the first successful attempt:
            - return self.direct_overrides[var](self, *args, **kw).
            - return self.direct_overrides_dynamic()[var](self, *args, **kw).
            - use self.load_fromfile.

        return the result (probably a numpy array, but not guaranteed).

        Examples:
            load Bx directly from a file
            load n for H+, using a different module which somehow gives nH+
                (PlasmaCalcs doesn't need to know where the value came from.)

        if used an override, instead of loading from file,
            set self._load_direct_used_override = var.
            Otherwise, set it to None.
            This might be used, e.g., to determine if the output came directly from a file or not.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        # check for overrides
        override = None
        try:  # direct_overrides  (check this before direct_overrides_dynamic.)
            override = self.direct_overrides[var]
        except KeyError:
            pass
        if override is None:  # direct_overrides_dynamic
            overrides_dynamic = self.direct_overrides_dynamic()
            try:
                override = overrides_dynamic[var]
            except KeyError:
                pass
        if override is not None:  # found an applicable override
            try:
                result = override(self, *args, **kw)
            except OverrideNotApplicableError:
                pass
            else:
                self._load_direct_used_override = var
                return result
        # no override found, or override not applicable.
        self._load_direct_used_override = None
        if not self.enable_fromfile:
            errmsg = (f'loading {var!r} currently requires load_fromfile, but enable_fromfile=False.\n'
                      'Maybe you forgot to self.set_var(), to define an override (see help(type(self).direct_overrides)),\n'
                      'or to restore enable_fromfile=True?')
            raise QuantCalcError(errmsg)
        return self.load_fromfile(var, *args, **kw)

    def load_fromfile(self, var, *args, **kw):
        '''load var directly from a file. Other methods should usually use load_direct, instead.
        the implementation here just raises LoadingNotImplementedError;
        subclasses should implement this method in order to load any values from files.
        '''
        raise LoadingNotImplementedError(f'{type(self).__name__}.load_fromfile')

    @property
    def direct_overrides(self):
        '''dict of {var: override} for all overrides of self which don't depend on behavior_attrs of self.
        For example, if user wants to set an override (or if setvars sets an override?), it will be here.
        See also: self.direct_overrides_dynamic().
        '''
        try:
            result = self._direct_overrides
        except AttributeError:
            result = dict()
            self._direct_overrides = result
        return result

    def direct_overrides_dynamic(self):
        '''returns dict of {var: override} for all overrides of self which depend on behavior_attrs of self.'''
        return dict()

    # # # LOAD FROM SETVARS OR CACHED VALUES # # #
    def get_set_or_cached(self, var):
        '''returns var if found in self.setvars or self.cache, with compatible behavior_attrs.
        otherwise, raise CacheNotApplicableError.

        if var is found in self.setvars and has relevant, but not matching behavior_attrs,
            self.load_across_dims will be used to load the value.
        '''
        behavior = self.behavior
        # -- from cache --
        # [TODO] <-- loading from self.cache.
        # -- from setvars --
        # [TODO][FIX] current implementation assumes user will not set var for multiple unit systems,
        #     when using ukey for some of those but not others. This might cause issues?
        relevance, quant = self.setvars.lookup(var, behavior)
        if not relevance:
            errmsg = f'var={var!r} not found in self.setvars, with compatible or relevant behavior_attrs.'
            raise CacheNotApplicableError(errmsg)
        if relevance is True:
            self.setvars.mark_used(var, quant)
            return quant.get_value(units_manager=getattr(self, 'u', None))
        # else: relevance is a list of subdims, i.e., iterable dimensions from self.behavior,
        #       for which there exists at least one point in the cache.
        def setvars_across_dims_loader(var):  # loader for a single dims point.
            result = self(var)
            if not isinstance(result, xr.DataArray):
                result = xr.DataArray(result)
                # [TODO] assign appropriate coords & attrs to this array.
            return result
        # [TODO][EFF] if loading 'var' would normally use partition_across_dim,
        #   use partition here too, but keep all "set vars" in a separate partition.
        #   (low priority; the existing implementation returns correct results,
        #    it's just unnecessarily slow compared to using partition logic,
        #    for long dims (e.g. >10 fluids) if only set a few of them (e.g. only set 1 or 2).)
        return self.load_across_dims(setvars_across_dims_loader, var, dims=relevance)

    # # # SETTING VARS # # #
    def set_var(self, var, value, behavior_attrs=None, forall=[], *, ukey=None, forced=False, **kw_using):
        '''set var in self. When later doing self(var) to get var, return the set value,
        but only if self.behavior is compatible with the relevant parts of self.behavior when var was set.

        This function will use, if it exists:
            self.KNOWN_SETTERS[var](self, value, behavior_attrs, forall=forall)
        Otherwise, calls:
            self.set_var_internal(var, value, self.behavior_attrs, forall=forall)
        
        var: str
            the var to set in self.
        value: number, xarray, iterable or 1D array, array with shape matching self.maindims_shape.
            the value to set var to.
            number --> set var to this number.
            xarray --> set var to this xarray.
            [TODO](not yet implemented) iterable or 1D array --> set var to these values along dim='testing'.
            [TODO](not yet implemented) array with shape matching self.maindims_shape --> set var to this array.
        behavior_attrs: None or list
            tells which attrs from self control behavior of the set var.
            The set var will only be retrieved when behavior_attrs of self are compatible.
                E.g. set_var('n', ['fluid', 'snap']) --> saves 'n' in cache with current fluid & snap.
                Will only load 'n' if self.fluid and self.snap == cached fluid and snap for 'n'.
            if var in self.KNOWN_SETTERS, cannot provide behavior_attrs here.
            else, use self.behavior_attrs if None.
        forall: list of strings
            if provided, tells which attrs of self do NOT control the behavior of the set var.
            E.g. forall=['snap'] --> 'snap' will NOT be included in behavior_attrs.
            (anything in behavior_attrs AND forall will be removed from the final behavior_attrs)
        ukey: None or str
            if provided, tells string to give to UnitsManager when converting value's units.
                When ukey is known, setting value in any unit system will enable to read it in all unit systems.
                E.g. set_var('n', 1e10, ..., ukey='n', units='si')
                --> self('n', units='raw') == self('n', units='si') * self.u('u', 'raw', convert_from='si')
            if not provided, value will be associated with current unit system;
                attempted to read value in any other unit system will not used the cached value set here.
                E.g. set_var('u', 1e10, ..., units='si')  # ukey not provided
                --> self('u', units='raw') --> uses self's other logic for getting 'u', not from setvars.
            note: if provided, 'units' will be added to behavior_attrs if not already in there.
        forced: bool, default True
            handles the case where self.KNOWN_SETTERS[var] doesn't exist. In that case...
            True --> set var in self, anyway.
            False --> crash; raise FormulaMissingError

        additional kwargs, if provided, go to self.using(**kw) during the operation.

        returns list of set quantities.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        with self.using(**kw_using):
            # use KNOWN_SETTERS[var] if possible
            try:
                setter = self.KNOWN_SETTERS[var]
            except KeyError:
                pass
            else:
                if behavior_attrs is not None:
                    errmsg = ('cannot provide behavior_attrs when var in KNOWN_SETTERS.\n'
                              '(If you *want* to bypass KNOWN_SETTERS, use set_var_internal instead.)\n'
                              f'got var={var!r}; behavior_attrs={behavior_attrs!r}.')
                    raise InputConflictError(errmsg)
                return setter(self, value, forall=forall)
            # default (KNOWN_SETTERS[var] not found)
            if forced:
                if behavior_attrs is None: behavior_attrs = self.behavior_attrs
                return self.set_var_internal(var, value, behavior_attrs, forall=forall, ukey=ukey)
            else:
                raise SetvarNotImplementedError(f'var={var!r} (not found in KNOWN_SETTERS), when forced=False.')

    setvar = alias('set_var')
    set = alias('set_var')

    def set_var_internal(self, var, value, behavior_attrs, forall=[], *, ukey=None):
        '''set var in self. KNOWN_SETTERS functions may wish to use this method.
        (KNOWN_SETTERS functions should NOT use self.set_var, to avoid recursion issue.)

        This function has the internal logic for self.set_var;
            set_var calls set_var_internal when self.KNOWN_SETTERS[var] not provided.
        
        var: str
            the var to set in self.
        value: number, xarray, iterable or 1D array, array with shape matching self.maindims_shape.
            the value to set var to. See help(self.set_var) for more info.
        behavior_attrs: list of strings
            the behavior attrs relevant to setting this var;
            getting var only gives value when current behavior attrs values are compatible with the cached ones.
        forall: list of strings
            if provided, tells which behavior attrs do NOT control the behavior of the set var.
            e.g. behavior_attrs=['snap', 'fluid'], forall=['snap'] --> use ['fluid'], only.
        ukey: None or str
            if provided, tells string to give to UnitsManager when converting value's units;
            when ukey is provided, can retrieve value in any unit system (probably 'si' or 'raw').
            when ukey not provided, if 'units' in used behavior attrs, can only retrieve value in that unit system.
        '''
        battrs = self._battrs_for_set_var_internal(behavior_attrs, forall, ukey=ukey)
        behavior = self.get_behavior(battrs)
        assign_attrs = dict(units=self.units) if 'units' in battrs else None
        # set the value.
        if getattr(value, 'ndim', 0) == 0:  # value is a number
            value_as_xarray = behavior.dims.assign_to(value, attrs=assign_attrs, name=var, overwrite=False)
            return self.setvars.append(var, value_as_xarray, behavior, ukey=ukey)
        elif isinstance(value, xr.DataArray):
            value_as_xarray = behavior.dims.assign_to(value, attrs=assign_attrs, name=var, overwrite=False)
            return self.setvars.append(var, value_as_xarray, behavior, ukey=ukey)
        else:  # other cases not yet implemented.
            errmsg = f'set_var for var={var!r} for value with ndim={value.ndim}, of type {type(value)}.'
            raise NotImplementedError(errmsg)

    def _battrs_for_set_var_internal(self, behavior_attrs, forall=[], *, ukey=None):
        '''returns behavior_attrs which will be used by set_var_internal, given these inputs.
        see help(self.set_var_internal) for details.
        '''
        # decide which behavior attrs to actually use, based on inputs.
        battrs = set(behavior_attrs) - set(forall)
        if ukey is not None:
            battrs.update({'units'})  # put 'units' in battrs if it's not already there.
        return battrs

    def unset_var(self, var, behavior_attrs=[], *, missing_ok=True, **kw_using):
        '''remove var from self.setvars (but only at values stored with relevant behavior).
        [TODO] define rules for which vars unset which other vars...
            e.g. for eppic right now, set_var('n') sets 'den' but not 'n';
            unset_var('n') unsets nothing... but should probably alias to unset_var('den').

        behavior_attrs: list of strings
            only remove cached values where self.behavior matches cached behavior for these attrs.
            if empty, remove all cached values for var, regardless of associated behavior.
        missing_ok: bool
            whether it is okay for there to be zero matching cached values for var.
            raise CacheNotApplicableError if missing_ok=False when there are no matching cached values.
        additional kwargs, if provided, go to self.using(**kw) during the operation.

        return list of CachedQuantity objects which were removed from self.setvars.
        '''
        with self.using(**kw_using):
            return self.unset_var_internal(var, behavior_attrs, missing_ok=missing_ok)

    unset = alias('unset_var')

    def unset_var_internal(self, var, behavior_attrs, forall=[], *, ukey=None, missing_ok=True):
        '''unset var from self.setvars.
        KNOWN_SETTERS functions may wish to use this method, to unset dependent values.
            E.g. if u depends on n, and n is changed, may wish to unset the value of u.

        behavior_attrs: list of strings
            the behavior attrs relevant to setting this var.
        forall: list of strings
            if provided, tells which behavior attrs to ignore when unsetting the var.
        ukey: None or string
            if provided, ignore 'units' behavior attr when unsetting the var
            (due to assuming that ukey was provided when setting the var,
            hence that the set var could be retrieved in any units system)
        missing_ok: bool
            whether it is okay for there to be zero matching cached values for var.
            raise CacheNotApplicableError if missing_ok=False when there are no matching cached values.

        return list of CachedQuantity objects which were removed from self.setvars.
        '''
        battrs = self._battrs_for_unset_var_internal(behavior_attrs, forall, ukey=ukey)
        behavior = self.get_behavior(battrs)
        return self.setvars.remove(var, behavior, missing_ok=missing_ok)

    def _battrs_for_unset_var_internal(self, behavior_attrs, forall=[], *, ukey=None):
        '''returns behavior_attrs which will be used by unset_var_internal, given these inputs.
        see help(self.unset_var_internal) for details.
        '''
        # decide which behavior attrs to actually use, based on inputs.
        battrs = behavior_attrs
        battrs = set(battrs) - set(forall)
        if ukey is not None:
            battrs = battrs - {'units'}  # remove 'units' from battrs if it's there.
        return battrs

    # # # HELP WITH QUANTS # # #
    # see quantities.help.py for details. Or, use QuantityLoader.help()

    # # # MISC... # # #
    # ATTACHING EXTRA COORDS # 
    cls_behavior_attrs.register('extra_coords', default={})
    extra_coords = simple_property('_extra_coords', setdefault=dict,
        doc='''dict of {coord_name: coord_value} to attach to outputs of self(var).
        Useful if planning to join the output of self(var) with output from a different QuantityLoader.
        E.g. self.extra_coords={'run': 'run 0'} and other.extra_coords={'run': 'run 1'},
            then xr.concat([self('n'), other('n')], 'run') gives 'n' from self AND other.
            (this is nice if self and other have same values for dims. Otherwise, might struggle.)''')

    def attach_extra_coords(self, arr):
        '''attach any self.extra_coords to array arr but only if it is an xarray.DataArray or xarray.Dataset'''
        extra_coords = self.extra_coords
        if (extra_coords is not None) and (len(extra_coords)>0) and isinstance(arr, (xr.DataArray, xr.Dataset)):
            arr = xarray_assign(arr, coords=self.extra_coords, overwrite=False)
            # ^ used overwrite=False to avoid overwriting any existing coords with the same name;
            # it's also more efficient and doesn't produce a new array, if all coords are already present.
        return arr

    # SCALING COORDS AT TOP LEVEL #
    cls_behavior_attrs.register('toplevel_scale_coords', default={})
    toplevel_scale_coords = simple_property('_toplevel_scale_coords', setdefault=dict,
        doc='''dict of {coord_name: coord_scaling} to apply to top-level outputs of self(var).
        (Never applies to internal calls of self(var), only applies at self.call_depth==1.)
        Useful if making plots and want to scale coords by some factor.
        E.g., self.toplevel_scale_coords = {'t': 1000} to convert s to ms.
        CAUTION: coord units labels will remain unaffected.''')

    def _apply_toplevel_scale_coords(self, arr):
        '''apply self.toplevel_scale_coords to arr, if nonempty, else return arr unchanged.'''
        scaling = self.toplevel_scale_coords
        assert self.call_depth == 1, f'called toplevel_scale_coords at 1!=call_depth(=={self.call_depth})'
        if (scaling is not None) and (len(scaling)>0) and isinstance(arr, (xr.DataArray, xr.Dataset)):
            arr = xarray_scale_coords(arr, scaling, missing_ok=True)
        return arr

    # WHETHER TO CRASH IF NAN TYPE #
    typevar_crash_if_nan = simple_property('_typevar_crash_if_nan', default=True,
        doc='''bool. whether to crash methods if typevar output would be 'nan'.
        False --> return NaN when typevar gives 'nan', instead of crashing.
        "typevar" here refers to any var used for checking which formula to use, from various options,
            e.g. 'ntype' in MhdMultifluidLoader or 'ionfrac_type' in MhdIonizationLoader.
        The relevant methods can check if self.typevar_crash_if_nan before returning a 'nan' result.''')

    def _handle_typevar_nan(self, *, errmsg=''):
        r'''crash with TypevarNanError if self.typevar_crash_if_nan, else return 'nan'.
        if crashing, use error message:
            errmsg + "\nTo return 'nan' instead of crashing, set self.typevar_crash_if_nan=False."
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if self.typevar_crash_if_nan:
            errmsg = errmsg + "\nTo return 'nan' instead of crashing, set self.typevar_crash_if_nan=False."
            raise TypevarNanError(errmsg)
        return 'nan'

    # # # COPYING # # #
    def copy(self):
        '''returns a deep copy of self.
        [TODO] implement something less hacky than using the pickle module?
        '''
        return copy_via_pickle(self)



"""
File Purpose: solving tfbi
"""

from ...tools import ImportFailed
try:
    import tfbi_theory as tt
except ImportError as err:
    tt = ImportFailed("tfbi_theory", err=err, locals=locals(), abbrv='tt')

from ...defaults import DEFAULTS
from ...errors import InputError
from ...tools import (
    format_docstring, alias, simple_property,
    UNSET,
    ProgressUpdater,
)

_paramdocs_tfbi_solving = {
    'solving_pattern': '''import tfbi_theory as tt
        import PlasmaCalcs as pc
        cc = ... # any PlasmaCalculator object from PlasmaCalcs.
        ds0 = cc.tfbi_ds()
        kp = tt.kPickerLowres(ds0)
        dsk = kp.get_ds()   # copy of ds0, but with ds['k'] = k from kPicker.
        drel = tt.TfbiDisprelC.from_ds(dsk)
        dsR = drel.solve()  # copy of dsk, but with ds['omega'] = solution to TFBI theory!''',
    'solving_pattern_notes': '''if cc has more than ~4 ions, you will want to drop some or group them somehow.
                e.g. for MhdMultifluidCalculator, before calling tfbi_ds():
                    cc.use_mix_heavy_ions(0.3, m_mean_mode='density')
            You can also pick ions directly during cc.tfbi_ds().
                e.g. for Bifrost chromosphere analysis, where n[He_II] < 1e-6 * ne, I use:
                    cc.tfbi_ds(ions=[i for i in cc.fluids if i.q==1 and i!='He_II'])
            After finishing solving, you might want to save the result,
                e.g. dsR.pc.save('filename') saves result to 'filename.pcxarr';
                can load it later via pc.xarray_load('filename.pcxarr').''',
}


''' --------------------- TfbiSolver --------------------- '''

@format_docstring(**_paramdocs_tfbi_solving,
                  max_num_ions=DEFAULTS.ADDONS.TFBI_MAX_NUM_IONS)
class TfbiSolver():
    '''high-level interface for solving TFBI theory across many physical parameters.

    Call TfbiSolver to solve TFBI.
    Example:
        import PlasmaCalcs as pc
        cc = pc.PlasmaCalculator(...)   # <-- your PlasmaCalculator of choice
        solver = pc.TfbiSolver(cc)
        solution = solver()             # alias: solver.solve()
    solution is an xarray.Dataset with all relevant quantities (see TfbiLoader.get_tfbi_all),
        and 'omega' telling roots with largest imaginary part.

    If you need more precise control over the solving process, use the pattern:
        {solving_pattern}

    TfbiSolver internally stores self.cc, ds0, kp, dsk, drel, and dsR,
        as defined by the pattern above.

    cc: PlasmaCalculator
        PlasmaCalculator object used to load the data.
        Should be a TfbiLoader subclass. (PlasmaCalculator satisfies this by default,
            assuming successful import SymSolver and import tfbi_theory.)
    ions: None or specifier of multiple fluids (e.g. slice, or list of strs)
        None --> use cc.fluids.ions()
        print warning if this specifies more than DEFAULTS.ADDONS.TFBI_MAX_NUM_IONS ions
            (default: {max_num_ions}), because then solving will be slow and may be inaccurate.
        ions are determined when called, not during __init__.
    kres: 'low', 'mid', or 'high'
        resolution in k-space. Tells which self.kPicker_cls to use.
        'low' --> tfbi_theory.kPickerLowres. Recommended if solving across many points.
        'mid' --> tfbi_theory.kPickerMidres. Recommended if solving across a few points.
        'high' --> tfbi_theory.kPickerHighres. Recommended if solving at only 1 point.
    mod, lmod, ang: UNSET or dict
        passed directly to kPicker if provided. Can specify k values other than the defaults.
        see help(self.kPicker_cls) for more details.
    tfbi_all: bool
        whether to compute all relevant tfbi vars, ds0 = cc('tfbi_all').
        False --> compute only the necessary vars, ds0 = cc('tfbi_inputs').
    drel_cls: None, str, or class
        tfbi_theory class to use for solving TFBI theory.
        None --> use self.drel_cls default: tt.TfbiDisprelC
        str --> use getattr(tt, drel_cls) to get the class.
    '''
    kPickerLowres_cls_name = 'kPickerLowres'  # used to make self.kPicker_cls if kres='low'
    kPickerMidres_cls_name = 'kPickerMidres'  # used to make self.kPicker_cls if kres='mid'
    kPickerHighres_cls_name = 'kPickerHighres'  # used to make self.kPicker_cls if kres='high'

    @property
    def kPicker_cls(self):
        '''kPicker class to use for choosing wavevectors to consider. Depends on self.kres:
        'low' --> tfbi_theory.kPickerLowres
        'mid' --> tfbi_theory.kPickerMidres
        'high' --> tfbi_theory.kPickerHighres
        '''
        if self.kres == 'low':
            return getattr(tt, self.kPickerLowres_cls_name)
        elif self.kres == 'mid':
            return getattr(tt, self.kPickerMidres_cls_name)
        elif self.kres == 'high':
            return getattr(tt, self.kPickerHighres_cls_name)
        else:
            raise InputError(f"Invalid kres. Expected 'low' or 'high'; got {self.kres!r}")

    drel_cls_name = 'TfbiDisprelC'  # used to make self.drel_cls.
    # (drel_cls = tt.TfbiDisprelC here would crash immediately if tt import failed.
    # By using drel_cls property, we avoid this crash until it's actually relevant.)
    drel_cls = simple_property('_drel_cls', setdefaultvia='_default_drel_cls',
        doc='''class to use for solving TFBI theory. Default: tt.TfbiDisprelC.''')
    def _default_drel_cls(self):
        '''default drel_cls. getattr(tt, self.drel_cls_name)'''
        return getattr(tt, self.drel_cls_name)

    def __init__(self, cc, ions=None, *, kres='low', mod=UNSET, lmod=UNSET, ang=UNSET, tfbi_all=True,
                 drel_cls=None):
        self.cc = cc
        self.ions = ions
        self.kres = kres
        self.mod = mod
        self.lmod = lmod
        self.ang = ang
        self.tfbi_all = tfbi_all
        if drel_cls is not None:
            if isinstance(drel_cls, str):
                self.drel_cls = getattr(tt, drel_cls)
            elif isinstance(drel_cls, type):
                self.drel_cls = drel_cls
            else:
                raise InputError(f"Invalid drel_cls. Expected None, str, or class; got {drel_cls!r}")

    @property
    def ions_explicit(self):
        '''list of ions from self.cc which would be used during self.cc.tfbi_ds()'''
        if self.ions is None:
            result = self.cc.fluids.ions()
        else:
            result = self.cc.fluids.get(self.ions)
        return result

    def _warn_if_too_many_ions(self):
        '''print warning if self.ions specifies too many ions.'''
        default_nmax = DEFAULTS.ADDONS.TFBI_MAX_NUM_IONS
        ions = self.ions_explicit
        if len(ions) > default_nmax:
            warnmsg = (f">>> Warning: TfbiSolver has {len(ions)} ions, which may be slow and inaccurate. <<<\n"
                       f"    (To avoid this warning, either use fewer ions, or "
                            f"increase DEFAULTS.ADDONS.TFBI_MAX_NUM_IONS (={default_nmax}, currently).)")
            print(warnmsg)

    __call__ = alias('solve')

    @format_docstring(**_paramdocs_tfbi_solving, sub_ntab=1)
    def solve(self, *, verbose=True, **kw_growth_root):
        '''solve TFBI theory. Assigns self.ds0, kp, dsk, drel, and dsR.
        Does not do any caching at this level; if the code crashes you will need to restart it completely.
        For more precise control including possibility for caching, use the pattern directly:
            {solving_pattern}

        verbose: bool
            whether to print progress updates (highly recommended).
        additional kwargs get passed directly to drel.solve().
            (options include: ncpu, ncoarse, careful)
        '''
        updater = ProgressUpdater(print_freq=0 if verbose else -1)  # print all statements if verbose, else nothing.
        self._warn_if_too_many_ions()
        # get ds0 -- manage progress updates here.
        ds0_updater = ProgressUpdater(print_freq=0 if verbose else -1)
        ds0_updater.print(f'getting ds0 via {type(self.cc).__name__}.tfbi_ds() ...')
        self.ds0 = self.cc.tfbi_ds(ions=self.ions, all=self.tfbi_all)
        ds0_updater.finalize(f'solver.ds0 = {type(self.cc).__name__}.tfbi_ds()', end='\n')
        # picking k is usually fast; no need for progress updates.
        kw_kPicker = dict(mod=self.mod, lmod=self.lmod, ang=self.ang)
        kw_kPicker = {k: v for k, v in kw_kPicker.items() if v is not UNSET}
        self.kp = self.kPicker_cls(self.ds0, **kw_kPicker)
        # follow the rest of the pattern. tt manages its own progress updates.
        self.dsk = self.kp.get_ds()
        self.drel = self.drel_cls.from_ds(self.dsk)
        self.dsR = self.drel.solve(verbose=verbose, **kw_growth_root)
        updater.print_clear(force=verbose)  # clear print statements on current line, if we're printing anything.
        updater.finalize(f'{type(self).__name__}.solve() (all steps from start to finish)', always=True)
        return self.dsR

    solution = alias('dsR', doc='''result of self.solve(); alias to self.dsR.''')
    solved = property(lambda self: hasattr(self, 'dsR'), doc='''tells whether self.solution exists''')

    def __repr__(self):
        contents = []
        cc_str = f'title={self.cc.title!r}' if hasattr(self.cc, 'title') else '...'
        contents.append(f'cc={type(self.cc).__name__}({cc_str})')
        if self.ions is not None:
            contents.append(f'ions={self.ions!r}')
        if isinstance(self.mod, dict) and len(self.mod) > 0:
            contents.append(f'mod={self.mod!r}')
        if isinstance(self.lmod, dict) and len(self.lmod) > 0:
            contents.append(f'lmod={self.lmod!r}')
        if isinstance(self.ang, dict) and len(self.ang) > 0:
            contents.append(f'ang={self.ang!r}')
        if self.tfbi_all != True:
            contents.append(f'tfbi_all={self.tfbi_all}')
        contents.append(f'solved={self.solved}')
        return f'{type(self).__name__}({", ".join(contents)})'



"""
File Purpose: tools related to documentation, e.g. docstrings
"""
import re
import types

from ..defaults import DEFAULTS

def format_docstring(*args__format, sub_indent=None, sub_ntab=None, **kw__format):
    '''returns a decorator of f which returns f, after updating f.__doc__ via f.__doc__.format(...)
    sub_indent: None or str
        if provided, indent all lines (after the first line) for each multiline string, before formatting.
    sub_ntab: int
        if provided when sub_indent is None, use sub_indent = sub_ntab * DEFAULTS.TAB.
    '''
    if sub_indent is None and sub_ntab is not None:
        sub_indent = DEFAULTS.TAB * sub_ntab
    if sub_indent is not None:
        args__format = [str(arg).replace('\n', '\n'+sub_indent) for arg in args__format]
        kw__format = {key: str(val).replace('\n', '\n'+sub_indent) for key, val in kw__format.items()}
    def return_f_after_formatting_docstring(f):
        f.__doc__ = f.__doc__.format(*args__format, **kw__format)
        return f
    return return_f_after_formatting_docstring

def docstring_but_strip_examples(f, *, strip_returns=True):
    '''return f.__doc__ but removing all lines after "Examples\n", if present.
    strip_returns: bool
        if True, also remove all lines after "Returns\n", if present.
    '''
    doc = f.__doc__
    if doc is None:
        return doc
    if 'Examples\n' in doc:
        doc = doc.split('Examples\n')[0].rstrip()
    if 'Returns\n' in doc:
        doc = doc.split('Returns\n')[0].rstrip()
    return doc


class DocstringInfo():
    r'''info about a docstring. Also, helpful methods for converting to sphinx format.

    docstring: str
        the original docstring. Stored at self.docstring, and never changed.
    kind: None, 'sphinx', or 'custom'
        what kind of docstring this is.
        None --> infer from docstring.
                'sphinx' if any line matches the KNOWN_SPHINX_PATTERN, else 'custom'.
                (e.g. if there's a :param...: line, then it's 'sphinx'.)
        'sphinx' --> docstring assumed to already follow all sphinx standards.
        'custom' --> docstring assumed to not follow sphinx standards. (See below)


    "Custom" format is used throughout most of PlasmaCalcs. Details:
        - first line is a description of the object. Never a param or other special line.
        - subsequent lines might continue the description.
        - line breaks are intended to be maintained when generating docs pages.
        - some lines tell params but don't specify :param:. Example:
            word: description with any number of words
        - after a param line, subsequent indented lines are a description of that param,
            until indentation level reverts back to same level as param line. Example:
            myparam1: description of myparam1
                continues description of myparam1

                also part of description of myparam1
            no longer part of description (might be a different param, or something else).
        - param "word" could instead be comma-separated words, but nothing fancier.
            E.g. "x, y: value" is treated as params, but "x space: value" is not.
            "Word" is defined to be a sequence of characters matching regex: \w
            Also, these can never be single-word params:
                'Example', 'Examples', 'Return', 'Returns'

        - currently, no other fancy-formatting is recognized.
            - (this might be updated eventually)
        - some patterns will be replaced:
            |x| --> \|x|   # to avoid sphinx interpreting as "substitution".
            `x` --> ``x``  # to tell sphinx to interpret as "code"
    If you don't like all the details of "custom" format, you can just use sphinx format.
    '''
    KNOWN_SPHINX_PATTERN = r'\s*(:param|:type|:return|:rtype|:raise).*:.*'
    PARAM_PATTERN = r'\s*((?!(?:Example|Return)s?)[\w]+(?:,\s*[\w]+)*)\s*:\s*(.+)'

    def __init__(self, docstring, *, kind=None):
        self.docstring = docstring  # original docstring, will never be changed
        self.lines = docstring.splitlines()
        self.kind = self._infer_kind(kind)
        # remove trailing blank lines.
        for i, l in enumerate(self.lines[::-1]):
            if not self.is_blank_line(l):
                break
        last_nonblank_line = len(self.lines) - i
        self.lines = self.lines[:last_nonblank_line]
        self.base_indent = self._infer_base_indent()

    def _infer_kind(self, kind):
        if kind is not None:
            return kind
        if any(re.fullmatch(self.KNOWN_SPHINX_PATTERN, l) for l in self.lines):
            return 'sphinx'
        else:
            return 'custom'

    def _infer_base_indent(self):
        '''return minimum indent across all lines except the first.
        if only nonblank line is the first line, return 0.
        '''
        result = None
        for l in self.lines[1:]:
            if not self.is_blank_line(l):
                indent = self.get_indent_n(l)
                if (result is None) or (indent < result):
                    result = indent
        if result is None:  # all nonblank lines have 0 indent
            result = 0
        return result

    @classmethod
    def from_obj(cls, obj):
        '''return DocstringInfo(obj.__doc__)'''
        return cls(obj.__doc__)

    @staticmethod
    def is_blank_line(line):
        '''returns whether this line is blank.'''
        return len(line.strip()) == 0

    @staticmethod
    def get_indent_n(line):
        '''returns number of leading spaces in line.'''
        return len(line) - len(line.lstrip())

    def nonblank_line_here_or_next(self, i):
        '''return self.lines[i] if nonblank, else first nonblank line after i.'''
        for j in range(i, len(self.lines)):
            if not self.is_blank_line(self.lines[j]):
                return self.lines[j]
        assert False, f'no nonblank lines at {i} or after'

    def param_ilines(self):
        '''return dict of {i: [iline of all description lines]} for param lines in docstring.
        param lines are those which match self.PARAM_PATTERN,
            AND are not indented underneath an existing param line.
            (also, line 0 is NEVER a param line, no matter what.)

        line number) example line:
        0) Example docstring
        1) param1: is a param line
        2)     description line
        3) param2: is another param line
        4)    description line
        5)        another description line
        6)    param 3: still not a param line (too indented)
        7) param4: param line again
        8) back to non-param line
        9)
        10) param5: param line again
        11)     description line again
        12) going to talk about a subset of parameters now:
        13)     subparam1: param line
        14)         description line
        15)     subparam2: param line
        16) 
        17) still non-paramline

        --> result would be:
            {1: [2], 3: [4, 5, 6], 7: [], 10: [11], 13: [14], 15: []}
        '''
        result = {}
        current_indent_n = None
        current_iparam = None
        for i, l in enumerate(self.lines):
            if i == 0:
                continue
            # check if sub-indented below a param_line.
            if current_indent_n is None:
                maybe_param_line = True
            else:  # currently considering maybe description lines.
                nonblank_l = self.nonblank_line_here_or_next(i)
                indent_n = self.get_indent_n(nonblank_l)
                if indent_n > current_indent_n:  # description line
                    maybe_param_line = False
                    result[current_iparam].append(i)
                else:  # no longer a description line
                    current_indent_n = None
                    current_iparam = None
                    maybe_param_line = True
            if maybe_param_line and re.fullmatch(self.PARAM_PATTERN, l):
                current_indent_n = self.get_indent_n(l)
                current_iparam = i
                result[i] = []
        return result

    def line_infos(self):
        '''return (i, linetype, indent_n, line_post_indent) for each line in docstring.

        i is line index.
        linetype is 'text', 'empty', 'param', or 'pdesc'.
            'text' --> misc text (not associated with a param)
            'empty' --> empty line
            'param' --> param line
            'pdesc' --> description line for param
        indent_n is number of spaces before text in line,
            used for prepending '| ' to ensure manual line breaks are respected in sphinx format.
            For blank lines, this is the indent_n of the next nonblank line.
            For text lines, this is the minimum indent across all lines (except the first line).
            For pdesc lines, this is the sub-indent level of the first desc line;
                subsequent desc lines have the same indent_n.
                (This makes the sphinx docs maintain the sub-indentation level as well!)
        line is all the text in the line after removing `indent_n` spaces.
        '''
        result = []
        param_ilines = self.param_ilines()
        current_indent_n = None
        i = 0
        result.append((i, 'text', 0, self.lines[i]))
        i = 1
        while i < len(self.lines):
            l = self.lines[i]
            if self.is_blank_line(l):
                nonblank_l = self.nonblank_line_here_or_next(i)
                indent_n = self.get_indent_n(nonblank_l)
                result.append((i, 'empty', indent_n, ''))
                i = i + 1
            elif i in param_ilines:
                current_indent_n = self.get_indent_n(l)
                result.append((i, 'param', current_indent_n, l[current_indent_n:]))
                if len(param_ilines[i]) >= 1:
                    desc0 = param_ilines[i][0]
                    desc0_nonblank = self.nonblank_line_here_or_next(desc0)
                    subindent_n = self.get_indent_n(desc0_nonblank)
                    for j in param_ilines[i]:
                        linej = self.lines[j]
                        if self.is_blank_line(linej):
                            result.append((j, 'pdesc', subindent_n, ''))
                        else:
                            result.append((j, 'pdesc', subindent_n, linej[subindent_n:]))
                    i = i + len(param_ilines[i])
                i = i + 1
            else:
                result.append((i, 'text', self.base_indent, l[self.base_indent:]))
                i = i + 1
        return result

    def _reconstruct(self):
        '''reconstruct docstring from self.line_infos().
        Mainly for debugging purposes.
        [TODO] result has extra (or not enough?) whitespace compared to docstring,
            but only differs in blank lines, so not a big deal.
        '''
        result = []
        for i, linetype, indent_n, line in self.line_infos():
            if linetype == 'empty':
                result.append('')
            elif linetype == 'text':
                result.append(' '*indent_n + line)
            elif linetype == 'param':
                result.append(' '*indent_n + line)
            elif linetype == 'pdesc':
                result.append(' '*indent_n + line)
            else:
                raise ValueError(f'unknown linetype: {linetype}')
        return '\n'.join(result)

    def to_sphinx(self):
        r'''returns self.docstring, converted to sphinx format.
        See help(type(self)) for more details.

        Equivalent: '\n'.join(self.to_sphinx_lines())
        '''
        return '\n'.join(self.to_sphinx_lines())
        
    def to_sphinx_lines(self):
        r'''returns lines of self.docstring, converted to sphinx format.
        See help(type(self)) for more details.

        Equivalent: self.to_sphinx().split('\n')
        '''
        if self.kind == 'sphinx':
            return self.lines
        elif self.kind == 'custom':
            return self._custom_to_sphinx_lines()
        else:
            raise ValueError(f'unknown kind: {self.kind}')

    def _custom_to_sphinx_lines(self):
        '''convert custom docstring to sphinx format.
        assumes but does not check self.kind=='custom'.
        '''
        result = []
        prev_linetype = None
        for i, linetype, indent_n, line in self.line_infos():
            # avoid sphinx interpreting line as "substitution"
            # (replace |x| with \|x|.)
            line = re.sub(r'[|]([^|]+)[|]', r'\|\g<1>|', line)
            # tell sphinx to interpret `x` as "code" (i.e. ``x``).
            line = re.sub(r'`([^`]+)`', r'``\g<1>``', line)
            # handle indentations properly
            if linetype == 'empty':
                result.append('')
            elif linetype == 'text':
                result.append(' '*indent_n + '| ' + line)
            elif linetype == 'param':
                # sphinx sometimes needs extra blank line between text & param
                # to render param appropriately.
                # (noticed this issue in particular when param indent > text indent.)
                if prev_linetype == 'text':
                    result.append('')
                result.append(' '*indent_n + line)
            elif linetype == 'pdesc':
                result.append(' '*indent_n + '| ' + line)
            else:
                raise ValueError(f'unknown linetype: {linetype}')
            prev_linetype = linetype
        return result

    # # # DISPLAY # # #
    def __repr__(self):
        l = self.nonblank_line_here_or_next(0)  # first nonblank line
        if len(self.lines) > 1:
            l = l + '...'
        if len(l) > 60:
            l = l[:60] + '...'
        return f'{type(self).__name__}({l})'


def sphinx_docstring(f):
    '''return sphinx docstring for this object.
    None if f.__doc__ is None or doesn't exist.
    Else, DocstringInfo(f.__doc__).to_sphinx()
    '''
    if getattr(f, '__doc__', None) is None:
        return None
    else:
        return DocstringInfo(f.__doc__).to_sphinx()

def list_objs(root, types_=(types.ModuleType, types.FunctionType, types.MethodType, type),
              *, recurse_on=(types.ModuleType, type), module_root=None, _seen=None):
    '''return list of all objects in root that are instances of any of types.

    root: where to start the search.
        result will only contain objects defined in root or sub-modules.
        (e.g. don't list numpy objects just because numpy was imported...)
    types_: tuple of types to check
        only list objects which are instances of any of these types.
        (trailing underscore avoids ambiguity with the `types` module.)
        All types here must have __module__ telling module name where they were defined,
            or __name__ if subclass of types.ModuleType.
    recurse_on: type of types
        for every obj which is an instance of recurse_on, also list_objs from that obj.
    module_root: None, str, or module
        only list objects defined in this module or sub-modules. E.g., 'PlasmaCalcs.tools'.
        None --> use root.__name__ (if root is module, else root.__module__
    _seen: internal parameter used for recursion;
        tracks which object ids have already been seen,
        to ensure each obj appears in result at most 1 time.

    Examples:
        import PlasmaCalcs as pc
        # list all modules, functions, methods, and classes in PlasmaCalcs
        ll = pc.list_objs(pc)
        # check that to_sphinx() works properly for all such objs in PlasmaCalcs
        for obj in ll:
            if obj.__doc__ is not None:
                pc.DocstringInfo(obj.__doc__).to_sphinx()
    '''
    seen = {} if _seen is None else _seen
    if module_root is None:
        module_root = root.__name__ if isinstance(root, types.ModuleType) else root.__module__
    for name in dir(root):
        obj = getattr(root, name, None)
        if id(obj) in seen:
            continue
        if isinstance(obj, types_):
            module_name = obj.__name__ if isinstance(obj, types.ModuleType) else obj.__module__
            if not module_name.startswith(module_root):
                continue
            seen[id(obj)] = obj
        if isinstance(obj, recurse_on):
            _list_here = list_objs(obj, types_=types_, recurse_on=recurse_on, # noqa: F841  #<comment for linter
                                    module_root=module_root, _seen=seen)
            # ^^ seen already tracks the result! _list_here saved for debugging purposes.
    return list(seen.values())



"""
File Purpose: densities for multifluid analysis of single-fluid mhd
"""
import xarray as xr

from .mhd_multifluid_ionization import MhdMultifluidIonizationLoader
from .mhd_fluids import Specie, IonMixture, ION_MIXTURE_SPECIES
from ..elements import Element
from ..mhd_eos_loader import MhdEosLoader
from ...defaults import DEFAULTS
from ...dimensions import SINGLE_FLUID, IONS
from ...errors import (
    FluidValueError, FluidKeyError,
    FormulaMissingError, LoadingNotImplementedError,
)
from ...tools import (
    alias, simple_property,
    xarray_promote_dim, xarray_sum,
)

class MhdMultifluidDensityLoader(MhdMultifluidIonizationLoader, MhdEosLoader):
    '''density quantities based on mhd single-fluid values, & inferred multifluid properties.'''

    # [TODO] (check is this actually correct?) inheritance notes:
    #   implementation here expects that EosLoader will be one of the parents.
    #   However, specifying it as a parent here directly will mess up the pattern of hookups
    #   which use, e.g. BifrostMultifluidCalculator(BifrostMultifluidStuff,
    #                           MhdMultifluidCalculator, BifrostCalculator),
    #   with BifrostCalculator inheriting from BifrostEosLoader which overrides some MhdEosLoader stuff.
    #   (if MhdMultifluidDensityLoader inherited from MhdEosLoader,
    #    then the BifrostMultifluidCalculator example above would use MhdEosLoader
    #    instead of BifrostEosLoader overrides.)


    # # # MISC DENSITY-RELATED VARS, OTHER THAN NUMBER DENSITY # # #

    @property
    def _has_fluid_with_density_weighted_mass(self):
        '''whether any fluid in self.fluid (not self.fluids) uses density-weighted mass.'''
        return any(isinstance(f, IonMixture) and f.m_mean_mode == 'density' for f in self.fluid_list())

    @known_var(load_across_dims=['fluid'],
               attr_deps=[('_has_fluid_with_density_weighted_mass', {True: ['n']})])
    def get_m(self):
        '''average mass of fluid particle.
        if SINGLE_FLUID, m computed as abundance-weighted average mass:
            m = self.elements.mtot() * (mass of 1 atomic mass unit).
            The "abundance-weighting" is as follows:
                m = sum_x(mx ax) / sum_x(ax), where ax = nx / nH, and x is any elem from self.elements.
                note: ax is related to abundance Ax via Ax = 12 + log10(ax).
            see help(self.elements.mtot) for more details, including a proof that mtot = rtot / ntot.
        if Element or Specie, return fluid.m, converted from [amu] to self.units unit system.
        if IonMixture, depends on m_mean_mode:
            'simple' --> mean mass of all ions in ion mixture
            'density' --> density-weighted average mass of all ions in mixture.
        '''
        f = self.fluid
        if f is SINGLE_FLUID:
            m = super().get_m()
        elif isinstance(f, (Element, Specie)):
            m = f.m * self.u('amu')
        else:
            assert isinstance(f, IonMixture), f'fluid of type {type(f)} not yet supported for get_m'
            if f.m_mean_mode == 'simple':
                m = f.m_mean * self.u('amu')
            elif f.m_mean_mode == 'density':
                with self.using(fluid=f.species):
                    nions = self('n')
                    mions = self('m')  # already in [self.units] units system.
                return xarray_sum(nions * mions, 'fluid') / xarray_sum(nions, 'fluid')
            else:
                raise LoadingNotImplementedError(f'get_m() for IonMixture with m_mean_mode={f.m_mean_mode!r}')
        return xr.DataArray(m, attrs=self.units_meta())

    @known_var(load_across_dims=['fluid'])  # [TODO] deps
    def get_r(self):
        '''mass density.
        if SINGLE_FLUID, r directly from Bifrost;
        if Element, r inferred from SINGLE_FLUID r and abundances;
        if Species or IonMixture, r = n * m.
        '''
        # [TODO][EFF] improve efficiency by allowing to group species, e.g. via Partition();
        #    self('n') * self('m') will get a good speedup if grouped instead of load_across_dims.
        f = self.fluid
        if f is SINGLE_FLUID:
            return super().get_r()
        elif isinstance(f, Element):
            return self('r_elem')
        elif isinstance(f, (Specie, IonMixture)):
            return self('n') * self('m')
        raise LoadingNotImplementedError(f'{type(self).__name__}.get_r() for fluid of type {type(f)}')


    # # # N_MODE DISPATCH / CODE ARCHITECTURE # # #
    cls_behavior_attrs.register('n_mode', default='best')
    cls_behavior_attrs.register('ne_mode', default=None)

    N_MODE_OPTIONS = {
        'best': '''use best mode available, based on fluid:
            electron --> 'table'
            Specie --> 'saha'.''',
        'elem': '''n for fluid's Element, from abundances and SINGLE_FLUID r.
            (crash if fluid.get_element() fails)''',
        'saha': '''from n_elem & saha ionization equation, assuming n=0 for twice+ ionized species.
            (crash if not Specie)''',
        'table': '''infer from EOS table, using SINGLE_FLUID r and e.
            (crash if not electron)''',
        'QN': '''sum of qi ni across self.fluids; getting 'ne' for saha via 'best' method.
            (crash if not electron)''',
        'QN_table': '''sum of qi ni across self.fluids; getting 'ne' for saha via 'table' method.
            (crash if not electron)''',
    }

    n_mode = simple_property('_n_mode', default='best', validate_from='N_MODE_OPTIONS',
        doc='''str. mode for getting Specie densities. (ignored if fluid is SINGLE_FLUID or an Element)
        see N_MODE_OPTIONS for details about available options.
        Note that you can always calculate n using a specific formula with the appropriate var,
            regardless of n_mode. E.g. n_saha will always load value from saha.
        Note: if ne_mode is not None, override n_mode with ne_mode when getting n for electrons.''')

    NE_MODE_OPTIONS = {
        None: '''use n_mode for electrons instead of ne_mode.''',
        'best': '''equivalent to table. Subclass might override.''',
        'table': '''infer from EOS table, using SINGLE_FLUID r and e.''',
        'QN': '''sum of qi ni across self.fluids; getting 'ne' for saha via 'best' method.''',
        'QN_table': '''sum of qi ni across self.fluids; getting 'ne' for saha via 'table' method.''',
    }

    _NE_MODE_INTERNAL = {
        'QN': 'best',
        'QN_table': 'table',
    }
    
    ne_mode = simple_property('_ne_mode', default=None, validate_from='NE_MODE_OPTIONS',
        doc='''None or str. mode for getting electron number density. See NE_MODE_OPTIONS for details.
        Setting self.ne_mode = '__internal__' will set ne_mode as if one layer deeper in the call;
            for many modes this just keeps ne_mode unchanged,
            but for QN modes this sets ne_mode to the internal value that will be used,
            e.g. if ne_mode was 'QN_table', setting ne_mode='__internal__' changes it to ne_mode='table'.''')
    _ne_mode_validate_and_fset = ne_mode.fset
    @ne_mode.setter
    def ne_mode(self, value):
        '''set self.ne_mode = value, but first check:
        if value == '__internal__', set to the internal value that will be used, 1 layer deeper in the call,
            i.e. self._NE_MODE_INTERNAL.get(self.ne_mode, self.ne_mode)
        '''
        if value == '__internal__':
            current_mode = self.ne_mode
            value = self._NE_MODE_INTERNAL.get(current_mode, current_mode)
        self._ne_mode_validate_and_fset(value)  # self._ne_mode = value, after validating value.

    ne_mode_explicit = alias('ne_mode', doc='''explicit ne_mode: ne_mode if not None, else n_mode.''')
    @ne_mode_explicit.getter
    def ne_mode_explicit(self):
        '''return ne_mode if set, else n_mode for electrons.'''
        ne_mode = self.ne_mode
        return self.n_mode if ne_mode is None else ne_mode

    # # # GENERIC NUMBER DENSITY # # #
    @known_var(load_across_dims=['fluid'])
    def get_ntype(self):
        '''ntype of self.fluid. Result depends on fluid as well as self.n_mode (and ne_mode if electron).
        See self.N_MODE_OPTIONS and self.NE_MODE_OPTIONS for options.
            'SINGLE_FLUID' <--> n for SINGLE_FLUID
            'elem' <--> n for Element
            'mixture' <--> sum of n for each ion in IonMixture
            'saha' <--> n from saha equation. (not available for electrons)
            'table' <--> n from EOS table. (only available for electrons)
            'QN_table' <--> n from sum of qi ni across self.fluids, with ne from table.
            'nan' <--> nan or crash, depending on self.typevar_crash_if_nan.
        '''
        result = 'nan'  # <-- nan unless set to something else below
        f = self.fluid
        if f is SINGLE_FLUID:
            result = 'SINGLE_FLUID'
        elif isinstance(f, Element):
            result = 'elem'
        elif isinstance(f, IonMixture):
            result = 'mixture'
        elif isinstance(f, Specie):
            electron = f.is_electron()
            mode = self.ne_mode_explicit if electron else self.n_mode
            if mode == 'elem' and f.element is not None:
                result = 'elem'
            elif (not electron) and ((mode == 'best') or (mode == 'saha')):
                result = 'saha'
            elif electron:
                if (mode == 'table') or (mode == 'best'):
                    result = 'table'
                elif (mode == 'QN_table') or (mode == 'QN'):
                    result = 'QN_table'
        if result == 'nan':
            self._handle_typevar_nan(errmsg=f"ntype, when fluid={self.fluid!r}, n_mode={self.n_mode!r}.")
        return xr.DataArray(result)

    # some (but not all) of the vars associated with ntypes.
    #   the types not included here are those with a more complicated dependency,
    #   e.g. when ntype=='QN_table' we need to do self('ne_QN', ne_mode='table').
    NTYPE_TO_VAR = {  # used in self.get_n. Subclass might copy this dict and add or adjust types.
        'elem': 'n_elem',
        'mixture': 'n_mixture',
        'saha': 'n_saha',
        'table': 'ne_fromtable',
        'nan': 'nan',
    }
    # non-simple deps based on ntype. Subclass might copy this dict and add or adjust types.
    #   "simple" dep if it is already described by NTYPE_TO_VAR.
    _NTYPE_TO_NONSIMPLE_DEPS = {
        'SINGLE_FLUID': ['SF_r', 'SF_m'],
        'QN_table': ['ne_QN', 'ne_fromtable'],
    }
    # all deps based on ntype. combination of NTYPE_TO_VAR and _NTYPE_TO_NONSIMPLE_DEPS.
    #   subclass might add new ntypes by adjusting those,
    #   or adjust the _NTYPE_TO_DEPS property to add an even-more-complicated dependency if needed.
    _NTYPE_TO_DEPS = property(lambda self: {**self.NTYPE_TO_VAR, **self._NTYPE_TO_NONSIMPLE_DEPS},
        doc='''all deps based on ntype. combination of NTYPE_TO_VAR and _NTYPE_TO_NONSIMPLE_DEPS.''')

    @known_var(partition_across_dim=('fluid', 'ntype'), partition_deps='_NTYPE_TO_DEPS')
    def get_n(self, *, ntype):
        '''number density. Formula depends on fluid:
        if SINGLE_FLUID, n = (r / m), from SINGLE_FLUID r & m.
            default m is the abundance-weighted average particle mass; see help(self.get_m) for details.
        if Element, n = (r / m), where
            r is inferred from abundances combined with SINGLE_FLUID r, and
            m is element particle mass (fluid.m)
        if Specie, n depends on ntype, determined from self.n_mode (and self.ne_mode, if electron);
            see help(self.get_ntype) for details.
        if IonMixture, n = sum of n for each Specie in the IonMixture.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if ntype in self.NTYPE_TO_VAR:
            var = self.NTYPE_TO_VAR[ntype]
            return self(var)
        elif ntype == 'SINGLE_FLUID':
            if self.current_n_fluid() > 1:
                raise LoadingNotImplementedError('[TODO] get_n with multiple SINGLE_FLUID...')
            with self.using(fluid=SINGLE_FLUID):
                return super().get_n()
        elif ntype == 'QN_table':
            return self('ne_QN', ne_mode='table')
        raise LoadingNotImplementedError(f'{type(self).__name__}.get_n() when ntype={ntype!r}.')


    # # # NTYPE: ELEM # # #

    @known_var(load_across_dims=['fluid'], aliases=['r_elem_per_rtot'])
    def get_rfrac_elem(self):
        '''mass density of element(s) for self.fluid, divided by total mass density.'''
        f = self.fluid
        if f is SINGLE_FLUID:
            return xr.DataArray(1)
        elif isinstance(f, (Element, Specie)):
            return xr.DataArray(f.get_element().r_per_nH() / self.elements.rtot_per_nH())
        else:
            raise NotImplementedError(f'fluid of type {type(f)} not yet supported for get_rfrac_elem')

    @known_var(load_across_dims=['fluid'], aliases=['n_elem_per_ntot'])
    def get_nfrac_elem(self):
        '''number density of element(s) for self.fluid, divided by total number density.'''
        f = self.fluid
        if f is SINGLE_FLUID:
            return xr.DataArray(1)
        elif isinstance(f, (Element, Specie)):
            return xr.DataArray(f.get_element().n_per_nH() / self.elements.ntot_per_nH())
        else:
            raise NotImplementedError(f'fluid of type {type(f)} not yet supported for get_nfrac_elem')

    @known_var(deps=['rfrac_elem', 'SF_r'])
    def get_r_elem(self):
        '''mass density of element(s) for self.fluid. r_elem = rfrac_elem * SF_r.'''
        return self('rfrac_elem') * self('SF_r')

    @known_var(deps=['nfrac_elem', 'SF_n'])
    def get_n_elem(self):
        '''number density of element(s) for self.fluid. n_elem = nfrac_elem * SF_n.'''
        return self('nfrac_elem') * self('SF_n')


    # # # NTYPE: MIXTURE # # #

    @known_var(load_across_dims=['fluid'], deps=[('n', {'fluid': ION_MIXTURE_SPECIES})])
    def get_n_mixture(self):
        '''sum of n for each ion in IonMixture.'''
        f = self.fluid
        if not isinstance(f, IonMixture):
            raise LoadingNotImplementedError(f'get_n_mixture() with fluid of type {type(f)}; expected IonMixture.')
        with self.using(fluid=f.species):
            nions = self('n')
        return xarray_sum(nions, 'fluid')


    # # # NTYPE: SAHA # # #
    # (n_saha defined in MhdMultifluidIonizationLoader)


    # # # NTYPE: ELECTRONS # # #
    @known_var(deps=['ntype'], ignores_dims=['fluid'])
    def get_ne_type(self):
        '''ne_type. Result depends on self.ne_mode.
        Possibilities include 'table', 'QN_table'. See help(self.get_ntype) for details.
        '''
        electron = self.fluids.get_electron()
        return self('ntype', fluid=electron)

    _NE_TYPE_TO_DEPS = {
        'table': ['ne_fromtable'],
        'QN_table': ['ne_QN', 'ne_fromtable'],
    }

    @known_var(value_deps=[('ne_type', '_NE_TYPE_TO_DEPS')])
    def get_ne(self):
        '''electron number density. Result depends on self.ne_mode.
        See self.NE_MODE_OPTIONS for details.
        '''
        kind = self('ne_type').item()
        if kind == 'table':
            result = self('ne_fromtable')
        elif kind == 'QN_table':
            result = self('ne_QN', ne_mode='table')
        elif kind == 'nan':
            return self._assign_electron_fluid_coord_if_unambiguous(self('nan'))
        else:
            errmsg = f'{type(self).__name__}.get_ne() when ne_mode={self.ne_mode_explicit!r}'
            raise LoadingNotImplementedError(errmsg)
        return result.assign_attrs(ne_type=kind)

    def _assign_electron_fluid_coord_if_unambiguous(self, array):
        '''return self.assign_fluid_coord(array, electron fluid).
        if self doesn't have exactly 1 electron fluid, don't assign coord.
        '''
        try:
            electron = self.fluids.get_electron()
        except FluidValueError:
            return array
        # else
        return self.assign_fluid_coord(array, electron, overwrite=True)

    @known_var(deps=['q', ('n', {'fluid': IONS, 'ne_mode': '__internal__'})], ignores_dims=['fluid'])
    def get_ne_QN(self):
        '''electron number density, assuming quasineutrality.
        result is sum_i qi ni / |qe|, with sum across all ions i in self.fluids.
        (Comes from assuming sum_s qs ns = 0, with sum across all species s in self.fluids.)
        '''
        ions = self.fluids.ions()
        if 'QN' in self.ne_mode_explicit:
            errmsg = (f"cannot get 'ne_QN' when ne_mode (={self.ne_mode_explicit!r}) still implies QN;\n"
                      "need a non-QN way to get ne for saha. See self.NE_MODE_OPTIONS for options.")
            raise FormulaMissingError(errmsg)
        ni = self('n', fluid=ions)  # <-- internally, ne for saha determined by self.ne_mode.
        Zi = self('q', fluid=ions) / self.u('qe')  # Zi = qi / |qe|
        result = Zi * ni
        result = xarray_promote_dim(result, 'fluid').sum('fluid')
        return self._assign_electron_fluid_coord_if_unambiguous(result)

    @known_var(deps=['SF_e', 'SF_r'], ignores_dims=['fluid'])
    def get_ne_fromtable(self):
        '''electron number density, from plugging r and e into eos tables (see self.tabin).'''
        result = super().get_ne_fromtable()  # see MhdEosLoader
        return self._assign_electron_fluid_coord_if_unambiguous(result)


    # # # --- SETTING VALUES; KNOWN SETTERS --- # # #
    # used when using set_var.

    @known_setter
    def set_n(self, value, **kw):
        '''set n to this value. n = number density.'''
        fluids = self.fluid_list()
        if all(f is SINGLE_FLUID for f in fluids):
            return super().set_n(value, **kw)
        relevant = ['snap', 'fluid']
        if not all((f is SINGLE_FLUID) or isinstance(f, Element) for f in fluids):
            relevant.append('n_mode')
            if not self.n_mode == 'elem':  # [TODO] any other cases where n doesn't depend on ne_mode?
                relevant.append('ne_mode')
        return self.set_var_internal('n', value, relevant, **kw, ukey='number_density')



"""
File Purpose: InstabilityQuantityLoader
"""
import xarray as xr

from .instability_data_tools import Pwl2FlatendFitter
from ..addon_tools import register_addon_loader
from ...dimensions import ELECTRON
from ...errors import InputError
from ...quantities import QuantityLoader
from ...tools import simple_property


''' --------------------- InstabilityQuantityLoader --------------------- '''

@register_addon_loader
class InstabilityQuantityLoader(QuantityLoader):
    '''quantities related to instability analysis of data.

    for analysis of theory, consider InstabilityCalculator instead.
    '''

    # # # FINDING / CHOOSING t_turb # # #

    cls_behavior_attrs.register('t_turb', default=None)
    t_turb = simple_property('_t_turb', default=None,
        doc='''None, or time of turbulent onset. some _turb quantities tell values only at t>=t_turb.
        Can be set to a DataArray, to test multiple possibilities at once.
        Units should be self.units. Changing self.units afterwards will NOT auto-update t_turb.''')

    @known_var
    def get_t_turb(self):
        '''time of turbulent onset. Equal to self.t_turb if set; Crash if not set.
        (here always converts result to DataArray if not DataArray already.)
        '''
        result = self.t_turb
        if result is None:
            errmsg = ('cannot get t_turb until it is set, via self.t_turb=value.\n'
                      'Also consider standard self.set_t_turb_... options, e.g. set_t_turb_00()')
            raise InputError(errmsg)
        if not isinstance(result, xr.DataArray):
            result = self.record_units(xr.DataArray(result))
        return result

    @known_pattern(r'pwl2_(.+)', deps=[0])
    def get_pwl2_var(self, var, *, _match=None):
        '''get pwl2_flatend fit to var across self.snap, evaluated at self.snap.
        (fail if self.snap is not a list of multiple snaps.)
        pwl2_flatend is a piecewise linear function with 2 pieces; final piece has slope=0.

        Equivalent: fitter=self(var).pc.pwl2_flatend_fitter('t'); fitter.fit(); fitter.eval()
        '''
        here, = _match.groups()
        arr = self(here)
        fitter = Pwl2FlatendFitter(arr, 't')
        fitter.fit()
        return fitter.eval()

    @known_pattern(r't_turb_from_pwl2_(.+)', deps=[0])
    def get_t_turb_from_pwl2_var(self, var, *, _match=None):
        '''t_turb from pwl2_flatend fit to var. Might want to do self.t_turb = result.
        result depends on self.blur_sigma.

        Equivalent: fitter=self('var').pc.pwl2_flatend_fitter('t'); fitter.fit(); fitter.get_xsat()

        Suggestion: t_turb_from_pwl2_ln_std_blur_deltafrac_n
        '''
        here, = _match.groups()
        arr = self(here)
        fitter = Pwl2FlatendFitter(arr, 't')
        fitter.fit()
        return fitter.get_xsat()

    @known_var(deps=['caches_ln_std_blur_deltafrac_n'])
    def get_tturbvar00(self):
        '''var used for t_turb '00' standard: caches_ln_std_blur_deltafrac_n,
            with fluid=electron, snap=None, blur_sigma=10.
        '''
        with self.using(t_turb=None, fluid=ELECTRON, snap=None, blur_sigma=10):
            return self('caches_ln_std_blur_deltafrac_n')

    def set_t_turb_00(self):
        '''set self.t_turb & return t_turb from '00' standard: t_turb_from_pwl2_ln_std_blur_deltafrac_n,
            with fluid=electron, snap=None, blur_sigma=10.
        Drops 'fluid', 'snap', and 't' coords from result.
        (internally, uses 'caches_ln_std_blur_deltafrac_n' with t_turb=None, to save time if recomputed later.)

        (00 standard should never change, it will always mean this!)
        '''
        val = self('t_turb_from_pwl2_tturbvar00')
        val = val.drop_vars(('fluid', 'snap', 't'), errors='ignore')
        self.t_turb = val
        return val

    def set_t_turb_10(self):
        '''set self.t_turb & return t_turb from '10' standard: t_turb_from_pwl2_ln_std_blur_deltafrac_n,
        for fluid=None, fitting across all snaps, and blur_sigma=10.
        Renames 'fluid' to 'tturb_fluid' in result. Drops 'snap' and 't' coords from result.
        (internally, uses 'caches_ln_std_blur_deltafrac_n' with t_turb=None, to save time if recomputed later.)

        (10 standard should never change, it will always mean this!)
        '''
        with self.using(t_turb=None, fluid=None, snap=None, blur_sigma=10):
            val = self('t_turb_from_pwl2_caches_ln_std_blur_deltafrac_n')
        val = val.rename({'fluid': 'tturb_fluid'}).drop_vars(('snap', 't'), errors='ignore')
        self.t_turb = val
        return val


    # # # COMPUTING VALUES BEFORE / AFTER t_turb # # #

    @known_pattern(r'tturb_(.+)', deps=['t_turb', 0])
    def get_tturb_var(self, var, *, _match=None):
        '''values of var at or after t_turb. Mask all values before t_turb.
        [TODO][EFF] wasted time computing values which will later be masked...
        '''
        here, = _match.groups()
        arr = self(here)
        t_turb = self('t_turb')
        return arr.where(arr['t'] >= t_turb)

    ## commented-out for now, because it feels bad to push t_final into playing a role in the results.
    ## maybe can think of a better thing to base the t_sureturb safety factor on.
    # cls_behavior_attrs.register('sureturb_quantile', default=0.3)
    # sureturb_quantile = simple_property('_sureturb_quantile', default=0.3,
    #     doc='''fraction of (t_final - t_turb) to add to t_turb to get t_sureturb.
    #     t_turb tells when turbulence probably starts.
    #     t_sureturb tells time, after which, values are definitely in the turbulent regime.
    #         --> to compute turbulent properties consider only t > t_sureturb.
    #     Example: t_turb=10, t_final=30, sureturb_quantile=0.3 --> t_sureturb=10+0.3*(30-10) = 16.''')

    @known_var(deps=['t_turb'])
    def get_t_sureturb(self):
        '''time, after which, values are definitely in the turbulent regime.
        Currently, just returns self('t_turb').
        Eventually might implement some safety factor since t_turb might not be exact.
        '''
        return self('t_turb')

    @known_pattern(r'sureturb_(.+)', deps=['t_sureturb', 0])
    def get_sureturb_var(self, var, *, _match=None):
        '''values of var at or after t_sureturb. Mask all values before t_sureturb.
        [TODO][EFF] wasted time computing values which will later be masked...
        '''
        here, = _match.groups()
        arr = self(here)
        t_sureturb = self('t_sureturb')
        return arr.where(arr['t'] >= t_sureturb)

    cls_behavior_attrs.register('surelin_quantile', default=0.2)
    surelin_quantile = simple_property('_surelin_quantile', default=0.2,
        doc='''fraction of t_turb which tells t_surelin.
        t_turb tells when turbulence probably starts.
        t_surelin tells time before which, values are definitely in the linear regime.
            --> to compute linear properties consider only t < t_surelin.
        Example: t_turb=10, surelin_quantile=0.2 --> t_surelin=0.2*10 = 2.

        See also: surelin_min_quantile.''')

    cls_behavior_attrs.register('surelin_min_quantile', default=0.05)
    surelin_min_quantile = simple_property('_surelin_min_quantile', default=0.05,
        doc='''fraction of t_turb which tells start time of "definitely linear regime".
        Use this to avoid including startup noise when computing linear properties.

        Example: t_turb=10, surelin_min_quantile=0.05 --> t_surelin=0.05*10 = 0.5.''')

    @known_var(deps=['t_turb'])
    def get_t_surelin(self):
        '''time before which, values are definitely in the linear regime.
        self('t_turb') * self.surelin_quantile.
        '''
        return self('t_turb') * self.surelin_quantile

    @known_var(deps=['t_turb'])
    def get_t_surelin_min(self):
        '''start time of "definitely linear regime".
        self('t_turb') * self.surelin_min_quantile.
        '''
        return self('t_turb') * self.surelin_min_quantile

    @known_pattern(r'surelin_(.+)', deps=['t_surelin', 0])
    def get_surelin_var(self, var, *, _match=None):
        '''values of var between (inclusive) t_surelin_min and t_surelin. Mask all other values.
        [TODO][EFF] wasted time computing values which will later be masked...
        '''
        here, = _match.groups()
        arr = self(here)
        t_surelin = self('t_surelin')
        t_surelin_min = self('t_surelin_min')
        return arr.where((arr['t'] >= t_surelin_min) & (arr['t'] <= t_surelin))

    @known_pattern(r'turblindiff_(.+)',
                   deps=[{0: 'meant_sureturb_{group0}'}, {0: 'meant_surelin_{group0}'}])
    def get_turblindiff_var(self, var, *, _match=None):
        '''meant_sureturb_var - meant_surelin_var.
        i.e., (time-averaged value in turbulent regime) minus (time-averaged value in linear regime).

        see also: werrturblindiff_var
        '''
        here, = _match.groups()
        sureturb = self(f'meant_sureturb_{here}')
        surelin = self(f'meant_surelin_{here}')
        return sureturb - surelin

    @known_pattern(r'turblindiffwerr_(.+)',
                     deps=[{0: '(werrmeant_sureturb_{group0})_werrsub_(werrmeant_surelin_{group0})'}])
    def get_turblindiffwerr_var(self, var, *, _match=None):
        '''werrmeant_sureturb_var - werrmeant_surelin_var.
        i.e., (time-averaged value in turbulent regime) minus (time-averaged value in linear regime),
        but result is a Dataset with 'mean' and 'std' data_vars,
        with 'std' coming from  "standard" error propagation formula assuming independent errors:
            std(A - B) = sqrt(std(A)**2 + std(B)**2).

        see also: turblindiff_var
        '''
        here, = _match.groups()
        return self(f'(werrmeant_sureturb_{here})_werrsub_(werrmeant_surelin_{here})')

    @known_pattern(r'turblindivwerr_(.+)',
                    deps=[{0: '(werrmeant_sureturb_{group0})_werrdiv_(werrmeant_surelin_{group0})'}])
    def get_turblindivwerr_var(self, var, *, _match=None):
        '''werrmeant_sureturb_var / werrmeant_surelin_var.
        i.e., (time-averaged value in turbulent regime) divided by (time-averaged value in linear regime),
        but result is a Dataset with 'mean' and 'std' data_vars,
        with 'std' coming from  "standard" error propagation formula assuming independent errors.
        '''
        here, = _match.groups()
        return self(f'(werrmeant_sureturb_{here})_werrdiv_(werrmeant_surelin_{here})')


    # # # ZEROTH ORDER NEUTRAL HEATING RATES # # #
    @known_var
    def get_dTndt_turb0_s_ds(self):
        '''dataset of contributions to neutral heating rate, based on 'mean' turbulent values,
        for each fluid (in self.fluid), due to collisions.

        Assumes u_n==0, and does not check.

        result has keys:
            'dTndt_u2': 2 m_n / (m_n + m_s) * nuns * [(m_s / (3 kB)) |u_s|^2]
            'dTndt_T': 2 m_n / (m_n + m_s) * nuns * [(T_s - T_n)]

        Except, use werrmeant_sureturb_u instead of u_s,
            and werrmeant_sureturb_T instead of T_s

        For more accurate computation, consider self('werrmeant_sureturb_dTndt_s')
        '''
        raise NotImplementedError('[TODO] -- maybe. Might not be needed though.')



"""
File Purpose: testing fft-related tools.
"""

import pytest

import PlasmaCalcs as pc


def test_fft_dimname():
    '''test FFTDimname class.'''
    # test creation
    pc.FFTDimname('x')   # should not crash
    pc.FFTDimname('x', rad=True)   # should not crash
    K_X = pc.DEFAULTS.FFT_FREQ_RAD_DIMNAMES.get('x', 'freqrad_x')
    assert K_X != 'freqrad_x'  # <-- if x removed from DEFAULTS.FFT_FREQ_RAD_DIMNAMES, can remove this line.
    # test from_post
    for postname, postrad in [('freq_x', False), ('freqrad_x', True), (K_X, True)]:
        dimname = pc.FFTDimname.from_post(postname, rad=None)
        assert dimname.name == 'x'
        assert dimname.rad == postrad
        pc.FFTDimname.from_post(postname, rad=postrad)  # should not raise error.
        with pytest.raises(pc.DimensionKeyError):
            pc.FFTDimname.from_post(postname, rad=not postrad)  # should raise error.
    # test __str__
    assert str(pc.FFTDimname('x')) == 'freq_x'
    assert str(pc.FFTDimname('x', rad=True)) == K_X
    # test implied_from
    for rad in (None, True, False):
        assert pc.FFTDimname.implied_from('x', ['x', 'y'], rad=rad) == pc.FFTDimname('x', rad=rad)
    assert pc.FFTDimname.implied_from(K_X, ['x', 'y']) == pc.FFTDimname('x', rad=True)
    assert pc.FFTDimname.implied_from('freq_x', ['x', 'y']) == pc.FFTDimname('x', rad=False)
    assert pc.FFTDimname.implied_from('x', [K_X, 'k_y'], post_fft=True) == pc.FFTDimname('x', rad=True)
    assert pc.FFTDimname.implied_from(K_X, [K_X, 'k_y'], post_fft=True) == pc.FFTDimname('x', rad=True)
    with pytest.raises(pc.DimensionKeyError):
        pc.FFTDimname.implied_from('freq_x', [K_X, 'k_y'], post_fft=True)
    with pytest.raises(pc.DimensionKeyError):
        pc.FFTDimname.implied_from('x', ['k_y', 't'])
    assert pc.FFTDimname.implied_from('x', ('fluid', 'x', 'y', 't')) == pc.FFTDimname('x')



"""
File Purpose: EppicSimInfoLoader
loading details about the simulation, not necessarily physics-based.
"""

import numpy as np
import xarray as xr

from .eppic_io_tools import n_mpi_processors
from ...errors import FluidKeyError, FileContentsMissingError
from ...quantities import QuantityLoader
from ...tools import (
    product,
    xarray_sum,
    slurm_options_here, slurm_option_here,
)

class EppicSimInfoLoader(QuantityLoader):
    '''simulation info for EppicCalculator.
    These are details about the simulation, not necessarily physics-based.
    E.g., number of particles per simulation cell, number of MPI processors used, simulation dx.

    (note that dx between cells in the output is actually dx_sim * nout_avg,
        since nout_avg is used to average over a few cells in space before providing the output.)
    '''

    # # # SLURM OPTIONS # # #

    @known_var
    def get_slurm_options(self):
        '''dataset of all slurm options from slurmfiles in self.dirname.
        crash if multiple slurmfiles in self.dirname with conflicting options.
        '''
        result = slurm_options_here(self.dirname)
        return xr.Dataset(result)

    @known_var(deps=['n_nodes', 'tasks_per_node'], aliases=['n_mpi_processors', 'n_mpi'])
    def get_n_processors(self):
        '''number of MPI processors used to run this run.
        From logfile if possible, else from slurm files.
        n_processors = n_nodes * tasks_per_node.
        '''
        try:
            result = n_mpi_processors(self.dirname)
        except (FileNotFoundError, AttributeError):
            result = self('n_nodes') * self('tasks_per_node')
        else:
            result = xr.DataArray(result)
        return result

    @known_var
    def get_n_nodes(self):
        '''number of nodes used to run this run. From slurm file.
        n_processors = n_nodes * tasks_per_node.
        '''
        # [TODO] support slurmfiles which specify n_processors instead of n_nodes.
        return xr.DataArray(slurm_option_here('--nodes', self.dirname))

    @known_var
    def get_tasks_per_node(self):
        '''number of processors per node. From slurm file.
        n_processors = n_nodes * tasks_per_node.
        '''
        slurm_options = slurm_options_here(self.dirname)
        if '--tasks-per-node' in slurm_options:
            result = xr.DataArray(slurm_options['--tasks-per-node'])
        elif '--total-tasks' in slurm_options and '--nodes' in slurm_options:
            result = slurm_options['--total-tasks'] // slurm_options['--nodes']
        else:
            errmsg = ('--tasks-per-node not found in slurm options,\n'
                      'and --total-tasks and --nodes not both present, either.')
            raise FileContentsMissingError(errmsg)
        return xr.DataArray(result)

    @known_var
    def get_time_requested(self):
        '''time limit for this run. From slurm file. Result is a string, like hh:mm:ss.
        For numeric results, see self('hours_requested'), minutes_requested, or seconds_requested.
        '''
        return xr.DataArray(slurm_option_here('--time', self.dirname))

    @known_pattern(r'(hours|minutes|mins|seconds)_requested', deps=['time_requested'])
    def get_time_requested_numeric(self, var, *, _match=None):
        '''time limit for this run, in hours, minutes, or seconds. From slurm file.
        Result is the total amount, e.g. '01:15:00' --> mins_requested=75.
        '''
        here, = _match.groups()
        time_str = self('time_requested', item=True)
        hours, minutes, seconds = [int(v) for v in time_str.split(':')]
        if here == 'hours':
            result = hours + minutes/60 + seconds/3600
        elif here in ('minutes', 'mins'):
            result = hours*60 + minutes + seconds/60
        elif here == 'seconds':
            result = hours*3600 + minutes*60 + seconds
        else:
            assert False, 'coding error if reached this line'
        return xr.DataArray(result)


    # # # EPPIC.I SIM INFO # # #

    @known_var
    def get_ndim_space(self):
        '''number of spatial dimensions in simulation. 2 or 3.'''
        return xr.DataArray(self.input_deck['ndim_space'])

    @known_var
    def get_nsubdomains(self):
        '''number of subdomains. nsubdomains from the input deck.
        Note: eppic runs require n_processors % nsubdomains == 0.
        '''
        return xr.DataArray(self.input_deck['nsubdomains'])

    @known_var(deps=['n_processors'])
    def get_npow2_processors(self):
        '''largest power of 2 which evenly divides n_processors.
        E.g. if n_processors = 56 * 8 == 7 * 4 * 8 == 7 * 2^5, result would be 2^5.
        '''
        n_processors = self('n_processors')
        largest_pow2_below = 2**(np.log2(n_processors).astype('int'))
        # (e.g. if n_processors = 7 * 2^5, largest_pow2_below == 2^2 * 2^5,
        #   because 2^2 < 7 but 2^3 > 7.)
        return np.gcd(n_processors, largest_pow2_below)

    @known_var(deps=['nsubdomains', 'tasks_per_node'])
    def get_min_n_nodes_given_nsubdomains(self):
        '''minimum number of nodes, given nsubdomains and tasks_per_node,
        to ensure n_processors % nsubdomains == 0,
            where n_processors = n_nodes * tasks_per_node.
        Equivalent: nsubdomains / gcd(nsubdomains, tasks_per_node)
        '''
        tasks_per_node = self('tasks_per_node')
        nsubdomains = self('nsubdomains')
        return nsubdomains // np.gcd(nsubdomains, tasks_per_node)

    @known_var
    def get_ncells_sim(self):
        '''number of gridcells from simulation (differs from output when nout_avg != 1).
        Scalar. E.g. Nx*Ny*Nz if 3D, Nx*Ny if 2D.
        '''
        nout_avg = self.input_deck['nout_avg']
        ncellsx = [self.input_deck.get_nspace(x)*nout_avg for x in self.maindims]
        return xr.DataArray(product(ncellsx))

    @known_var(load_across_dims=['component'])
    def get_ds_sim(self):
        '''grid spacing (of simulation). vector(ds), e.g. [dx, dy, dz]. Depends on self.component.
        ds_sim = (dx, dy, dz) from input deck (not divided by nout_avg)
        '''
        x = str(self.component)
        dx = self.input_deck[f'd{x}'] * self.u('length')
        return xr.DataArray(dx, attrs=dict(units=self.units))

    @known_var
    def get_dt_sim(self):
        '''time spacing (of simulation). Time between iterations (not between snapshots)'''
        return xr.DataArray(self.input_deck['dt'], attrs=dict(units=self.units))

    def npd_for_fluid(self, fluid):
        '''return the npd for this fluid.
        This is equivalent to fluid['npd'] when it is provided,
            otherwise determined by the appropriate alternative (npcelld, nptotd, or nptotcelld).
        This method is implemented for the calculator rather than the fluid, 
            because fluid doesn't know the possibly-required global values (ncells and/or n_processors).

        result will always be converted to int, since npd is an integer.
        '''
        key = None
        for key in ('npd', 'npcelld', 'nptotd', 'nptotcelld'):
            if key in fluid.keys():
                break
        else:  # didn't break
            raise FluidKeyError(f'fluid {fluid} does not have npd, npcelld, nptotd, or nptotcelld.')
        if key == 'npd':
            return int(fluid['npd'])
        elif key == 'npcelld':
            return int(fluid['npcelld'] * self('ncells_sim', item=True))
        elif key == 'nptotd':
            return int(fluid['nptotd'] / self('n_processors', item=True))
        elif key == 'nptotcelld':
            return int(fluid['nptotcelld'] * self('ncells_sim/n_processors', item=True))
        else:
            assert False, "coding error if reached this line"

    @known_var(load_across_dims=['fluid'])
    def get_npd(self):
        '''number of PIC particles in each distribution.
        This is equivalent to fluid['npd'] when it is provided,
            otherwise determined by the appropriate alternative (npcelld, nptotd, or nptotcelld).
        '''
        return xr.DataArray(self.npd_for_fluid(self.fluid))

    @known_var(deps=['npd', 'ncells_sim'])
    def get_npcelld(self):
        '''number of PIC particles per simulation cell.'''
        return self('npd') / self('ncells_sim')

    @known_var(deps=['npd', 'n_processors'])
    def get_nptotd(self):
        '''number of PIC particles (total across all processors).'''
        return self('npd') * self('n_processors')

    @known_var(deps=['npd', 'n_processors', 'ncells_sim'])
    def get_nptotcelld(self):
        '''number of PIC particles per simulation cell (total across all processors).'''
        return self('npd') * self('n_processors') / self('ncells_sim')

    @known_var(load_across_dims=['fluid'])
    def get_subcycle(self):
        '''subcycling factor (for each fluid in self.fluid).
        (If subcycle not provided for a distribution, assume it implies subcycle=1).
        '''
        return xr.DataArray(self.fluid.get('subcycle', 1))

    @known_var(deps=['nptotcelld', 'subcycle'], reduces_dims=['fluid'])
    def get_nptotcell(self):
        '''number of particles per cell for all fluids combined, appropriately weighted considering subcycling.
        Equivalent: sum_fluids_(nptotcelld/subcycle)
        '''
        with self.using(fluid=None):
            return xarray_sum(self('nptotcelld') / self('subcycle'), 'fluid', missing_dims='ignore')

    @known_var
    def get_nit_since_prev(self):
        '''return number of timesteps since previous snapshot in self.
        when determining previous snapshot, ignore any where snap.file_snap(self) is MISSING_SNAP.
        return inf when no previous snap, e.g. at snap=0.
        '''
        if not all(s.exists_for(self) for s in self.iter_snap()):
            return self('_nit_since_prev_simple')
        snap2i = {}
        for i, snap in enumerate(self.snaps):
            snap2i[id(snap)] = i
        result = []
        for i, snap in enumerate(self.iter_snap()):
            iprev = snap2i[id(snap)] - 1
            if iprev < 0:
                result.append(np.inf)
            else:
                it_here = int(snap.file_s(self))
                it_prev = int(self.snaps[iprev].file_s(self))
                result.append(it_here - it_prev)
        if self.snap_is_iterable():  # result is 1D
            return xr.DataArray(result, dims='snap', coords=dict(snap=self.snap))
        else:  # result is 0D
            return xr.DataArray(result, coords=dict(snap=self.snap))

    @known_var(load_across_dims=['snap'])
    def get__nit_since_prev_simple(self):
        '''naive implementation of nit_since_prev. Makes fewer assumptions, but can be slow.
        (e.g., took 0.8 seconds for 800 snaps. Compare to 0.02 seconds via nit_since_prev.)
        nit_since_prev should dispatch to this method if any of the assumptions fail.
        '''
        snap_here = self.snap  # this will be a single snap thanks to load_across_dims=['snap']
        if not snap_here.exists_for(self):
            return xr.DataArray(self.snap_dim.NAN)
        # get previous existing snap.
        snapi = int(self.snap) if self.snap==self.snaps[int(self.snap)] else self.snaps.index(self.snap)
        prev_snap_i = snapi - 1
        while prev_snap_i >= 0:
            prev_snap = self.snaps[prev_snap_i]
            if prev_snap.exists_for(self):
                break
            else:
                prev_snap_i -= 1
        else:  # didn't break; no previous snap found.
            return xr.DataArray(np.inf)
        # <-- at this point, prev_snap is the previous existing snap, and snap_here is the current snap.
        it_here = int(snap_here.file_s(self))
        it_prev = int(prev_snap.file_s(self))
        return xr.DataArray(it_here - it_prev)


    # # # CPU COST # # #
    @known_var(deps=['n_processors', 'timer2seconds', 'timers', 'nit_since_prev'])
    def get_cpu_seconds_per_timestep(self):
        '''cpu seconds per timestep, for each timer in timers.
        n_processors * timer2seconds * timers / nit_since_prev

        see also: cpu_seconds_per_ct, cpu_seconds_per_pct
        '''
        numer_coeff = self('n_processors') * self('timer2seconds')
        return (numer_coeff / self('nit_since_prev')) * self('timers')

    @known_var(deps=['cpu_seconds_per_timestep', 'ncells_sim'])
    def get_cpu_seconds_per_ct(self):
        '''cpu seconds per cell per timestep, for each timer in timers.
        n_processors * timer2seconds * timers / (ncells_sim * nit_since_prev)

        see also: cpu_seconds_per_timestep, cpu_seconds_per_pct
        '''
        return self('cpu_seconds_per_timestep') / self('ncells_sim')

    @known_var(deps=['nptotcell', 'cpu_seconds_per_ct'])
    def get_cpu_seconds_per_pct(self):
        '''cpu seconds per particle, per cell, per timestep, for each timer in timers.
        n_processors * timer2seconds * timers / (nptotcell * ncells_sim * nit_since_prev)
        (note, nptotcell accounts for subcycling. See self.help('nptotcell') for details.)

        see also: cpu_seconds_per_timestep, cpu_seconds_per_ct
        '''
        return self('cpu_seconds_per_ct') / self('nptotcell')

    @known_var(deps=['cpu_seconds_per_ct', 'nptotcell'])
    def get_cpu_seconds_per_each(self):
        '''cpu seconds per (particle, per) cell, per timestep, as relevant to each timer in timers.

        cpu_seconds_per_pct for definitely scaling with nptotcell:
            - vadv time
            - xadv time
            - charge
        cpu_seconds_per_pct for timers probably scaling with nptotcell:
            - output
        cpu_seconds_per_ct for timers definitely NOT scaling with nptotcell:
            - collect
            - efield
        
        EXCLUDE timers combining times above:
            - Wall Clock
            - Sys Clock
        EXCLUDE timers not handled here:
            - fluid
        '''
        per_ct = self('cpu_seconds_per_ct')
        result = per_ct.drop_vars(('Wall Clock', 'Sys Clock', 'fluid'))
        pct_vars = ['vadv time', 'xadv time', 'charge', 'output']
        return result.assign(per_ct[pct_vars] / self('nptotcell'))

    @known_var(deps=['nptotcell'])
    def get_cpu_per_each2ct(self):
        '''conversion factors: cpu_per_each2ct * cpu_per_each == cpu_per_ct.
        Result is a dataset with timers as data vars.
        These convert from (per particle per cell per timestep) to (per cell per timestep):
            'vadv time', 'xadv time', 'charge', 'output'.
            (I.e., for these timers: result == nptotcell)
        These do not convert at all:
            'collect', 'efield'
            (I.e., for these timers: result == 1)
        Other timers' conversions are not included.
        '''
        result = {}
        nptotcell = self('nptotcell')
        for timer in ['vadv time', 'xadv time', 'charge', 'output']:
            result[timer] = nptotcell
        for timer in ['collect', 'efield']:
            result[timer] = 1
        return xr.Dataset(result)

    @known_var(deps=['cpu_per_each2ct', 'ncells_sim'])
    def get_cpu_per_each2timestep(self):
        '''conversion factors: cpu_per_each2timestep * cpu_per_each == cpu_per_timestep.
        Result is a dataset with timers as data vars.
        These convert from (per particle per cell per timestep) to (per timestep):
            'vadv time', 'xadv time', 'charge', 'output'.
            (I.e., for these timers: result == nptotcell * ncells_sim)
        These convert from (per cell per timestep) to (per timestep):
            'collect', 'efield'
            (I.e. for these timers: result == ncells_sim)
        Other timers' conversions are not included.
        '''
        return self('cpu_per_each2ct') * self('ncells_sim')



"""
File Purpose: InstabilityCalculator specialized in computing EPPIC values.
Input is an xarray.Dataset.
Knows how to write input deck files for eppic, based on values.
"""

from .eppic_dist_inputs_loader import EppicDistInputsLoader
from .eppici_maker import EppiciMaker
from .eppic_glob_inputs_loader import EppicGlobInputsLoader
from .eppic_instability_plotters import EppicInstabilityPlotterManager
from .eppic_safety_info_loader import EppicSafetyInfoLoader
from ....errors import InputConflictError
from ....other_calculators import InstabilityCalculator
from ....tools import UNSET, format_docstring
from ....defaults import DEFAULTS

@format_docstring(NDIM_SPACE=DEFAULTS.EPPIC.NDIM_SPACE, 
                  NPTOTCELLD0=DEFAULTS.EPPIC.NPTOTCELLD0,
                  DIST_DEFAULTS=DEFAULTS.EPPIC.DIST_DEFAULTS,
                  GLOB_DEFAULTS=DEFAULTS.EPPIC.GLOB_DEFAULTS)
class EppicInstabilityCalculator(EppiciMaker, EppicDistInputsLoader, EppicGlobInputsLoader,
                                 EppicSafetyInfoLoader, EppicInstabilityPlotterManager,
                                 InstabilityCalculator):
    '''InstabilityCalculator specialized in computing EPPIC values,
    and with helpful method to make an EPPIC input deck in zeroth order equilibrium.

    ds: xarray.Dataset
        should probably have all relevant base vars (& simple derived vars).
        Possibilities include:
            ds, m, q, gamma, n, u, T, nusj, nusn, E, B, E_un0,
            m_n or m_neutral, n_n or n_neutral, u_n or u_neutral, T_n or T_neutral.

    ndim_space: UNSET or int
        number of spatial dimensions (probably 2 or 3).
        UNSET --> ds['ndim_space'] if provided, else DEFAULTS.EPPIC.NDIM_SPACE (default={NDIM_SPACE}).
    nptotcelld0: UNSET or int
        nptotcelld for the unsubcycled distribution (not necessarily dist 0).
        UNSET --> ds['nptotcelld0'] if provided, else DEFAULTS.EPPIC.NPTOTCELLD0 (default={NPTOTCELLD0}).
        (raise InputConflictError if ds['nptotcelld0'] conflicts with value input here.)
    glob_vals: UNSET or dict of values
        can specify non-default values for global parameters.
        Recognizes all keys (will add as a data_var in self.ds),
        but some defaults are already defined in DEFAULTS.EPPIC.GLOB_DEFAULTS:
            {GLOB_DEFAULTS}.
    dist_vals: UNSET or dict of (value or xarray.DataArray with 'fluid' dimension)
        can specify non-default values for misc. distribution parameters.
        Recognizes all keys (will add as a data_var in self.ds),
        but some defaults are already defined in DEFAULTS.EPPIC.DIST_DEFAULTS:
            {DIST_DEFAULTS}.
        To specify new value but same value for each dist, use non-DataArray value.
        To specify different values for each dist, use {{key: DataArray with 'fluid' dimension}}.
        This is really just an alternative to manually adding these values to ds;
            any values provided directly in ds will not be overridden.
    '''
    # fluid_list_cls = EppicDistList   # using this would cause crash when using non-EppicDist Fluids,
    # jfluid_list_cls = EppicNeutralList   # which is not actually desireable...

    def __init__(self, ds, *, ndim_space=UNSET, nptotcelld0=UNSET,
                 glob_vals=UNSET, dist_vals=UNSET, **kw_super):
        self._kw_input = dict(ndim_space=ndim_space, nptotcelld0=nptotcelld0,
                              dist_vals=dist_vals, glob_vals=glob_vals)
        ds = self.assign_inputs(ds)
        super().__init__(ds, **kw_super)  # includes: self.ds = self.assign_bases(ds)

    def assign_inputs(self, ds):
        '''return a copy of ds, after assigning relevant init inputs from self.
        Currently, checks:
            - nptotcelld0
            - dist_vals
        '''
        # ndim_space
        ndim_space_in = self._kw_input['ndim_space']
        ndim_space_ds = ds.get('ndim_space', None)
        if ndim_space_ds is not None and ndim_space_in is not UNSET:
            if ndim_space_ds != ndim_space_in:
                raise InputConflictError(f"ndim_space conflict: {ndim_space_ds} != {ndim_space_in}")
        if ndim_space_ds is None:
            if ndim_space_in is UNSET:
                ndim_space_in = DEFAULTS.EPPIC.NDIM_SPACE
            ds = ds.assign(ndim_space=ndim_space_in)
        # nptotcelld0  # [TODO] encapsulate instead of repeating code from ndim_space^?
        nptotcelld0_in = self._kw_input['nptotcelld0']
        nptotcelld0_ds = ds.get('nptotcelld0', None)
        if nptotcelld0_ds is not None and nptotcelld0_in is not UNSET:
            if nptotcelld0_ds != nptotcelld0_in:
                raise InputConflictError(f"nptotcelld0 conflict: {nptotcelld0_ds} != {nptotcelld0_in}")
        if nptotcelld0_ds is None:
            if nptotcelld0_in is UNSET:
                nptotcelld0_in = DEFAULTS.EPPIC.NPTOTCELLD0
            ds = ds.assign(nptotcelld0=nptotcelld0_in)
        # dist_vals
        dv_use = DEFAULTS.EPPIC.DIST_DEFAULTS.copy()
        dv_in = self._kw_input['dist_vals']
        if dv_in is UNSET: dv_in = {}
        dv_use.update(dv_in)
        for var, value in dv_use.items():
            if var in ds:
                if var in dv_in:
                    raise InputConflictError(f'ds already has {var!r}; cannot also provide in dist_vals.')
                # else, pass. Don't override ds value with value from DIST_DEFAULTS.
            else:
                ds = ds.assign({var: value})
        # glob_vals   # [TODO]: encapsulate instead of repeating code from dist_vals^?
        gv_use = DEFAULTS.EPPIC.GLOB_DEFAULTS.copy()
        gv_in = self._kw_input['glob_vals']
        if gv_in is UNSET: gv_in = {}
        gv_use.update(gv_in)
        for var, value in gv_use.items():
            if var in ds:
                if var in gv_in:
                    raise InputConflictError(f'ds already has {var!r}; cannot also provide in glob_vals.')
                # else, pass. Don't override ds value with value from GLOB_DEFAULTS.
            else:
                ds = ds.assign({var: value})
        return ds



"""
File Purpose: basic test for muram hookup in PlasmaCalcs.
These tests require the Header.NNN file(s), but not any simulation output data.
"""
import os
import pytest

import numpy as np

import PlasmaCalcs as pc

HERE = os.path.dirname(__file__)
RUNDIR = os.path.join(HERE, 'test_muram_nodata')


''' --------------------- basics --------------------- '''

def get_muram_calculator(**kw_init):
    '''make a muram calculator'''
    with pc.InDir(RUNDIR):
        mc = pc.MuramCalculator(**kw_init)
    return mc

def get_muram_multifluid_calculator(**kw_init):
    '''make a muram multifluid calculator'''
    with pc.InDir(RUNDIR):
        mm = pc.MuramMultifluidCalculator(**kw_init)
    return mm

def test_make_muram_calculator():
    '''ensure can make MuramCalculator and MuramMultifluidCalculator objects.'''
    mc = get_muram_calculator()
    mm = get_muram_multifluid_calculator()


''' --------------------- misc --------------------- '''

def test_muram_params():
    '''test some stuff about muram params'''
    mc = get_muram_calculator()
    mm = get_muram_multifluid_calculator()
    EXPECTED = {   # hard-coded some expected values
        'order': np.array([1, 2, 0]),
        'N0': 1152,
        'N1': 512,
        'N2': 256,
        'Nx': 512,
        'Ny': 256,
        'Nz': 1152,
        'dx': 19200000.0,
        'dy': 19200000.0,
        'dz': 6400000.0,
        't': 49926.96,
        'snap_s': '557956',
    }
    for param, value in EXPECTED.items():
        assert np.all(mc.params[param] == value)
        assert np.all(mm.params[param] == value)

def test_muram_mixture():
    '''test some stuff about muram elements and species.'''
    mc = get_muram_calculator()
    mm = get_muram_multifluid_calculator()
    # check elements:
    ELEMENTS = ['H', 'He', 'C', 'N', 'O',
                'Ne', 'Na', 'Mg', 'Al', 'Si',
                'S', 'K', 'Ca', 'Cr', 'Fe', 'Ni']
    assert len(mc.elements) == len(ELEMENTS)
    assert len(mm.elements) == len(ELEMENTS)
    assert all(e == el for e, el in zip(mc.elements, ELEMENTS))
    assert all(e == el for e, el in zip(mm.elements, ELEMENTS))
    # check mass:
    assert mc('m') == mc.elements.mtot() * mc.u('amu')
    assert mm('SF_m') == mm.elements.mtot() * mm.u('amu') == mc('m')
    assert mm('SINGLE_FLUID_m') == mm('m', fluid=pc.SINGLE_FLUID) == mm('SF_m')
    # check species details:
    assert len(mm.fluid) == len(mm.fluids) # fluid=None by default
    assert len(mm.fluids.ions()) == len(mm.elements)  # once-ionized ion of each element
    assert all(f.q==1 for f in mm.fluids.ions())
    assert len(mm.fluids) == len(mm.elements) + 2  # 2= 1 electron, 1 neutral
    assert mm.fluid[0] == mm.fluids.get_electron()
    # check can get mass of species:
    mions = mm('m', fluid=mm.fluids.ions())
    assert mions.identical(mm('m', fluid=pc.IONS))
    assert np.all(mions == [el.m * mm.u('amu') for el in mm.elements])
    mall = mm('m')
    assert mall.isel(fluid=0) == mm.u('me')
    # check can get charge of species:
    qall = mm('q')
    assert qall.isel(fluid=0) == -mm.u('qe')
    assert qall.isel(fluid=1) == 0  # fluid[1] is neutral in this case.
    #    ^ [TODO] maybe don't need to impose this; it's okay if muram changes this fact...
    assert np.all(qall.isel(fluid=slice(2, None)) == mm.u('qe'))
    # ensure SF_m doesn't change if altering mm.fluids (bug fixed on 2025-04-16)
    i_helium = [f.i for f in mm.fluids if getattr(f, 'element', None)=='He']
    mm_fluids_orig = mm.fluids
    mm.fluids = mm.fluids.without_i(i_helium)
    mm.fluid = None  # <-- restore result.fluid to "all fluids"
    assert len(mm.fluids) < len(mm_fluids_orig)  # i.e. this is a nontrivial test...
    assert mm('SF_m').item() == mc('SF_m').item()

def test_match_var_tree():
    '''test that match_var_tree works for a few vars'''
    mc = get_muram_calculator()
    mm = get_muram_multifluid_calculator()
    assert mm.tree == mm.match_var_tree
    mm.eos_mode = 'aux'
    # test across a variety of vars, just ensuring tree doesn't crash:
    for var in ['m', 'B', 'u', 'T', 'nusn', 'skappa', 'E', 'E_parmod_B']:
        mc.tree(var)
        mm.tree(var)
    # test tree('n') across a variety of fluid, n_mode, eos_mode, and ne_mode options,
    #   because the multifluid number densities logic can get complicated...
    # (takes a few (~3) seconds. There are >500 combinations here...)
    mm.typevar_crash_if_nan = False
    for fluid in [None, pc.NEUTRAL, pc.ELECTRON]:
        for n_mode in mm.N_MODE_OPTIONS:
            for ne_mode in mm.NE_MODE_OPTIONS:
                for eos_mode in mm.EOS_MODE_OPTIONS:
                    info = dict(fluid=fluid, n_mode=n_mode, ne_mode=ne_mode, eos_mode=eos_mode)
                    mm.tree('n', **info)
    # test tree crashes appropriately when typevar_crash_if_nan=True
    mm.typevar_crash_if_nan = True
    with pytest.raises(pc.TypevarNanError):
        mm.tree('n', fluid=pc.ELECTRON, n_mode='elem')  # can't get n_elem for electron!

def test_mix_heavy_ions():
    '''test that we can use mix_heavy_ions() for MuramMultifluidCalculator.'''
    mm = get_muram_multifluid_calculator()
    mm.use_mix_heavy_ions()
    mix1 = mm.use_mix_heavy_ions(m_tol=0.1)
    mix3 = mm.use_mix_heavy_ions(m_tol=0.3)
    assert mix1 != mix3
    assert any(isinstance(f, pc.IonMixture) for f in mix1)
    ion_mixes = mix3.ion_mixtures()
    mm.fluid = pc.ION_MIXTURES
    assert mm.fluid == ion_mixes
    # check that tree('n') doesn't crash, even when using IonMixtures:
    for n_mode in mm.N_MODE_OPTIONS:
        for eos_mode in mm.EOS_MODE_OPTIONS:
            mm.tree('n')
    # check that pc.ION_MIXTURE_SPECIES doesn't crash:
    mm.fluid = pc.ION_MIXTURE_SPECIES
